import argparse
import json
import time

import numpy as np

from delft.textClassification import Classifier
from delft.textClassification.models import architectures
from delft.textClassification.reader import load_dataseer_corpus_csv
from delft.textClassification.reader import vectorize as vectorizer
from delft.utilities.Utilities import split_data_and_labels

"""
    Classifier for deciding if a sentence introduce a dataset or not, and prediction of the
    dataset type.
"""


def configure(architecture):
    # default RNN model parameters
    batch_size = 200
    maxlen = 300
    patience = 5
    early_stop = True
    max_epoch = 50
    learning_rate = 0.001

    # default transformer model parameters
    if architecture == "bert":
        batch_size = 16
        early_stop = False
        max_epoch = 5
        maxlen = 300
        learning_rate = 2e-5

    return batch_size, maxlen, patience, early_stop, max_epoch, learning_rate


def train(embeddings_name, fold_count, architecture="gru", transformer=None, cascaded=False):
    print("loading binary dataset type corpus...")
    xtr, y, _, _, list_classes, _, _ = load_dataseer_corpus_csv("data/textClassification/dataseer/all-binary.csv")

    model_name = "dataseer-binary_" + architecture
    class_weights = None

    batch_size, maxlen, patience, early_stop, max_epoch, learning_rate = configure(architecture)

    model = Classifier(
        model_name,
        architecture=architecture,
        list_classes=list_classes,
        max_epoch=max_epoch,
        fold_number=fold_count,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        patience=patience,
        early_stop=early_stop,
        class_weights=class_weights,
        transformer_name=transformer,
        learning_rate=learning_rate,
    )

    if fold_count == 1:
        model.train(xtr, y)
    else:
        model.train_nfold(xtr, y)
    # saving the model
    model.save()

    print("loading reuse dataset type corpus...")
    xtr, y, _, _, list_classes, _, _ = load_dataseer_corpus_csv("data/textClassification/dataseer/all-reuse.csv")

    model_name = "dataseer-reuse_" + architecture
    class_weights = {0: 1.5, 1: 1.0}

    model = Classifier(
        model_name,
        architecture=architecture,
        list_classes=list_classes,
        max_epoch=max_epoch,
        fold_number=fold_count,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        patience=patience,
        early_stop=early_stop,
        class_weights=class_weights,
        transformer_name=transformer,
        learning_rate=learning_rate,
    )

    if fold_count == 1:
        model.train(xtr, y)
    else:
        model.train_nfold(xtr, y)
    # saving the model
    model.save()

    print("loading first-level dataset type corpus...")
    xtr, y, _, _, list_classes, _, _ = load_dataseer_corpus_csv("data/textClassification/dataseer/all-multilevel.csv")

    model_name = "dataseer-first_" + architecture

    class_weights = None

    model = Classifier(
        model_name,
        architecture=architecture,
        list_classes=list_classes,
        max_epoch=max_epoch,
        fold_number=fold_count,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        patience=patience,
        early_stop=early_stop,
        class_weights=class_weights,
        transformer_name=transformer,
        learning_rate=learning_rate,
    )

    if fold_count == 1:
        model.train(xtr, y)
    else:
        model.train_nfold(xtr, y)
    # saving the model
    model.save()

    """
    print('training second-level dataset subtype corpus...')
    xtr, y1, y2, _, list_classes, list_subclasses, _ = load_dataseer_corpus_csv("data/textClassification/dataseer/all-multilevel.csv")
    # aggregate by class, we will have one training set per class
    datatypes_y = {}
    datatypes_xtr = {}
    datatypes_list_subclasses = {}
    for i in range(0,len(xtr)):
        datatype = y1[i]
        datasubtype = y2[i]
        if datatype in datatypes_y:
            datatypes_y[datatype].append(datasubtype)
            datatypes_xtr[datatype].append(xtr[i])
            if not y2[i] in datatypes_list_subclasses[datatype]:
                datatypes_list_subclasses[datatype].append(y2[i])
        else:
            datatypes_y[datatype] = []
            datatypes_y[datatype].append(datasubtype)
            datatypes_xtr[datatype] = []
            datatypes_xtr[datatype].append(xtr[i])
            datatypes_list_subclasses[datatype] = []
            datatypes_list_subclasses[datatype].append(y2[i])

    for the_class in list_classes:
        print('training', the_class)

        model_name = 'dataseer-' + the_class + "_" + architecture

        model = Classifier(model_name, architecture=architecture, list_classes=datatypes_list_subclasses[the_class], max_epoch=max_epoch,
            fold_number=fold_count, patience=patience, use_roc_auc=True, embeddings_name=embeddings_name,
            batch_size=batch_size, class_weights=class_weights, early_stop=early_stop, transformer_name=transformer)

        if fold_count == 1:
            model.train(datatypes_xtr[the_class], datatypes_y[the_class])
        else:
            model.train_nfold(datatypes_xtr[the_class], datatypes_y[the_class])
        # saving the model
        model.save()
    """


def train_and_eval(embeddings_name=None, fold_count=1, architecture="gru", transformer=None, cascaded=False):
    if cascaded:
        return train_eval_cascaded(embeddings_name, fold_count, architecture=architecture, transformer=transformer)

    # classifier for deciding if we have a dataset or not in a sentence
    train_and_eval_binary(embeddings_name, fold_count, architecture=architecture, transformer=transformer)

    # classifier for deciding if the introduced dataset is a reuse of an existing one or is a new dataset
    # train_and_eval_reuse(embeddings_name, fold_count, architecture=architecture, transformer=transformer)

    # classifier for first level data type hierarchy
    # train_and_eval_primary(embeddings_name, fold_count, architecture=architecture, transformer=transformer)

    # classifier for second level data type hierarchy (subtypes)
    # train_and_eval_secondary(embeddings_name, fold_count, architecture=architecture, transformer=transformer)


def train_and_eval_binary(embeddings_name, fold_count, architecture="gru", transformer=None):
    print("loading dataset type corpus...")
    # xtr, y, _, _, list_classes, _, _ = load_dataseer_corpus_csv("data/textClassification/dataseer/all-binary.csv")
    xtr, y, _, _, list_classes, _, _ = load_dataseer_corpus_csv("data/textClassification/dataseer/phase1-2-binary.csv")
    # xtr, y, _, _, list_classes, _, _ = load_dataseer_corpus_csv("data/textClassification/dataseer/phase3-binary.csv")

    # distinct values of classes
    print(list_classes)
    print(len(list_classes), "classes")

    print(len(xtr), "texts")
    print(len(y), "classifications")

    class_weights = None

    batch_size, maxlen, patience, early_stop, max_epoch, learning_rate = configure(architecture)

    model = Classifier(
        "dataseer-binary_" + architecture,
        architecture=architecture,
        list_classes=list_classes,
        max_epoch=max_epoch,
        fold_number=fold_count,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        patience=patience,
        early_stop=early_stop,
        class_weights=class_weights,
        transformer_name=transformer,
        learning_rate=learning_rate,
    )

    # segment train and eval sets
    x_train, y_train, x_test, y_test = split_data_and_labels(xtr, y, 0.9)

    # x_train = xtr
    # y_train = y

    # print('loading dataset type corpus...')
    # x_test, y_test, _, _, _, _, _ = load_dataseer_corpus_csv("data/textClassification/dataseer/phase3-binary.csv")

    # _, _, x_test, y_test = split_data_and_labels(x_test, y_test, 0.9)

    # distinct values of classes
    # print(list_classes)
    # print(len(list_classes), "classes")

    print(len(xtr), "texts")
    print(len(y), "classifications")

    print(len(x_train), "train texts")
    print(len(y_train), "train classes")

    print(len(x_test), "eval texts")
    print(len(y_test), "eval classes")

    if fold_count == 1:
        model.train(x_train, y_train)
    else:
        model.train_nfold(x_train, y_train)
    model.eval(x_test, y_test)

    # saving the model
    model.save()


def train_and_eval_reuse(embeddings_name, fold_count, architecture="gru", transformer=None):
    print("loading dataset type corpus...")
    xtr, y, _, _, list_classes, _, _ = load_dataseer_corpus_csv("data/textClassification/dataseer/all-reuse.csv")

    # distinct values of classes
    print(list_classes)
    print(len(list_classes), "classes")

    print(len(xtr), "texts")
    print(len(y), "classes")

    batch_size, maxlen, patience, early_stop, max_epoch, learning_rate = configure(architecture)

    class_weights = {0: 1.5, 1: 1.0}

    model = Classifier(
        "dataseer-reuse_" + architecture,
        architecture=architecture,
        list_classes=list_classes,
        max_epoch=max_epoch,
        fold_number=fold_count,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        patience=patience,
        early_stop=early_stop,
        class_weights=class_weights,
        transformer_name=transformer,
        learning_rate=learning_rate,
    )

    # segment train and eval sets
    x_train, y_train, x_test, y_test = split_data_and_labels(xtr, y, 0.9)

    print(len(x_train), "train texts")
    print(len(y_train), "train classes")

    print(len(x_test), "eval texts")
    print(len(y_test), "eval classes")

    if fold_count == 1:
        model.train(x_train, y_train)
    else:
        model.train_nfold(x_train, y_train)
    model.eval(x_test, y_test)

    # saving the model
    model.save()


def train_and_eval_primary(embeddings_name, fold_count, architecture="gru", transformer=None):
    print("loading dataset type corpus...")
    xtr, y, _, _, list_classes, _, _ = load_dataseer_corpus_csv("data/textClassification/dataseer/all-multilevel.csv")

    # distinct values of classes
    print(list_classes)
    print(len(list_classes), "classes")

    print(len(xtr), "texts")
    print(len(y), "classes")

    class_weights = None
    batch_size, maxlen, patience, early_stop, max_epoch, learning_rate = configure(architecture)

    model = Classifier(
        "dataseer-first_" + architecture,
        architecture=architecture,
        list_classes=list_classes,
        max_epoch=max_epoch,
        fold_number=fold_count,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        patience=patience,
        early_stop=early_stop,
        class_weights=class_weights,
        transformer_name=transformer,
        learning_rate=learning_rate,
    )

    # segment train and eval sets
    x_train, y_train, x_test, y_test = split_data_and_labels(xtr, y, 0.9)

    print(len(x_train), "train texts")
    print(len(y_train), "train classes")

    print(len(x_test), "eval texts")
    print(len(y_test), "eval classes")

    if fold_count == 1:
        model.train(x_train, y_train)
    else:
        model.train_nfold(x_train, y_train)
    model.eval(x_test, y_test)

    # saving the model
    model.save()


def train_and_eval_secondary(embeddings_name, fold_count, architecture="gru", transformer=None):
    print("training second-level dataset subtype corpus...")
    xtr, y1, y2, _, list_classes, list_subclasses, _ = load_dataseer_corpus_csv(
        "data/textClassification/dataseer/all-multilevel.csv"
    )
    # aggregate by class, we will have one training set per class

    print(list_classes)
    print(list_subclasses)
    print(len(list_classes), "classes")
    print(len(list_subclasses), "sub-classes")

    class_weights = None
    batch_size, maxlen, patience, early_stop, max_epoch, learning_rate = configure(architecture)

    datatypes_y = {}
    datatypes_xtr = {}
    datatypes_list_subclasses = {}
    for i in range(0, len(xtr)):
        # print(np.where(y2[i] == 1))
        ind1 = np.where(y1[i] == 1)[0][0]
        ind2 = np.where(y2[i] == 1)[0][0]
        # print(ind2)
        datatype = list_classes[ind1]
        datasubtype = list_subclasses[ind2]
        # print(str(xtr[i]), datatype, datasubtype)
        if datatype in datatypes_y:
            datatypes_y[datatype].append(datasubtype)
            datatypes_xtr[datatype].append(xtr[i])
            if datasubtype not in datatypes_list_subclasses[datatype]:
                datatypes_list_subclasses[datatype].append(datasubtype)
        else:
            datatypes_y[datatype] = []
            datatypes_y[datatype].append(datasubtype)
            datatypes_xtr[datatype] = []
            datatypes_xtr[datatype].append(xtr[i])
            datatypes_list_subclasses[datatype] = []
            datatypes_list_subclasses[datatype].append(datasubtype)

    print(datatypes_list_subclasses)

    for the_class in list_classes:
        print("\ntraining", the_class)
        if the_class not in datatypes_list_subclasses:
            print("no subclass for", the_class)
            continue

        if len(datatypes_list_subclasses[the_class]) <= 1:
            print("only one subclass for", the_class)
            continue

        if len(datatypes_list_subclasses[the_class]) == 2 and "nan" in datatypes_list_subclasses[the_class]:
            continue

        if the_class == "Protein Data":
            continue

        print("subtypes to be classified:", datatypes_list_subclasses[the_class])

        model_name = "dataseer-" + the_class + "_" + architecture

        model = Classifier(
            model_name,
            architecture=architecture,
            list_classes=datatypes_list_subclasses[the_class],
            max_epoch=max_epoch,
            fold_number=fold_count,
            use_roc_auc=True,
            embeddings_name=embeddings_name,
            batch_size=batch_size,
            maxlen=maxlen,
            patience=patience,
            early_stop=early_stop,
            class_weights=class_weights,
            transformer_name=transformer,
            learning_rate=learning_rate,
        )

        # we need to vectorize the y according to the actual list of classes
        local_y = []
        for the_y in datatypes_y[the_class]:
            the_ind = datatypes_list_subclasses[the_class].index(the_y)
            local_y.append(vectorizer(the_ind, len(datatypes_list_subclasses[the_class])))

        # segment train and eval sets
        x_train, y_train, x_test, y_test = split_data_and_labels(
            np.asarray(datatypes_xtr[the_class]), np.asarray(local_y), 0.9
        )

        if fold_count == 1:
            model.train(x_train, y_train)
        else:
            model.train_nfold(x_train, y_train)
        model.eval(x_test, y_test)
        # saving the model
        model.save()


def classify(texts, output_format, architecture="gru"):
    """
    Classify a list of texts with an existing model
    """
    # load model
    model = Classifier("dataseer-binary_" + architecture)
    model.load()
    start_time = time.time()
    result = model.predict(texts, output_format)
    runtime = round(time.time() - start_time, 3)
    if output_format == "json":
        result["runtime"] = runtime
    else:
        print("runtime: %s seconds " % (runtime))
    return result


def train_eval_cascaded(embeddings_name, fold_count, architecture="gru", transformer=None):
    # general setting of parameters
    class_weights = None
    batch_size, maxlen, patience, early_stop, max_epoch, learning_rate = configure(architecture)

    # first binary classifier: dataset or no_dataset
    xtr, y, _, _, list_classes, _, _ = load_dataseer_corpus_csv("data/textClassification/dataseer/all-binary.csv")

    print(list_classes)

    model_binary = Classifier(
        "dataseer-binary_" + architecture,
        architecture=architecture,
        list_classes=list_classes,
        max_epoch=max_epoch,
        fold_number=fold_count,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        patience=patience,
        early_stop=early_stop,
        class_weights=class_weights,
        transformer=transformer,
        learning_rate=learning_rate,
    )

    # segment train and eval sets
    x_train, y_train, x_test, y_test = split_data_and_labels(xtr, y, 0.9)

    if fold_count == 1:
        model_binary.train(x_train, y_train)
    else:
        model_binary.train_nfold(x_train, y_train)
    model_binary.eval(x_test, y_test)

    x_test_binary = x_test

    # second, the first level datatype taxonomy for sentences classified as dataset
    xtr, y_classes, y_subclasses, y_leafclasses, list_classes, list_subclasses, list_leaf_classes = (
        load_dataseer_corpus_csv("data/textClassification/dataseer/all-multilevel.csv")
    )
    # ignore the no_dataset, ignore the first eval set, build first level classifier

    ind = list_classes.index("no_dataset")
    to_remove = vectorizer(ind, len(list_classes))

    x_train, y_train = filter_exclude_class(xtr, y_classes, to_remove)
    y_train2 = np.zeros(shape=(len(y_train), len(list_classes) - 1))
    for i in range(0, len(y_train)):
        y_train2[i] = np.delete(y_train[i], ind)
    y_train = y_train2

    list_classes.remove("no_dataset")

    model_first = Classifier(
        "dataseer-first_" + architecture,
        architecture=architecture,
        list_classes=list_classes,
        max_epoch=100,
        fold_number=fold_count,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        patience=patience,
        early_stop=early_stop,
        class_weights=class_weights,
        transformer=transformer,
        learning_rate=learning_rate,
    )

    if fold_count == 1:
        model_first.train(x_train, y_train)
    else:
        model_first.train_nfold(x_train, y_train)
    model_first.eval(x_test, y_test)

    # eval by cascading
    result_binary = model_binary.predict(x_test_binary, output_format="default")
    model_first.predict(x_test, output_format="default")

    # select sequences classified as dataset
    result_intermediate = np.asarray([np.argmax(line) for line in result_binary])

    def vectorize(index, size):
        result = np.zeros(size)
        if index < size:
            result[index] = 1
        return result

    result_binary = np.array([vectorize(xi, len(list_classes)) for xi in result_intermediate])


def filter_exclude_class(xtr, y_classes, the_class):
    # apply the filter
    new_xtr = []
    new_classes = []
    for i in range(0, len(y_classes)):
        if not np.array_equal(y_classes[i], the_class):
            new_xtr.append(xtr[i])
            new_classes.append(y_classes[i])

    return np.array(new_xtr), np.array(new_classes)


def build_prior_class_distribution():
    """
    Inject count from the training data to the classification taxonomy of data types
    """
    _, y_classes, y_subclasses, y_leafclasses, list_classes, list_subclasses, list_leaf_classes = (
        load_dataseer_corpus_csv("data/textClassification/dataseer/all-multilevel.csv")
    )

    with open("data/textClassification/dataseer/DataTypes.json") as json_file:
        distribution = json.load(json_file)

    # init count everywhere
    for key1 in distribution:
        print(key1)
        if type(distribution[key1]) is dict:
            distribution[key1]["count"] = 0
            for key2 in distribution[key1]:
                print(key1, key2)
                if type(distribution[key1][key2]) is dict:
                    distribution[key1][key2]["count"] = 0
                    for key3 in distribution[key1][key2]:
                        print(key1, key2, key3)
                        if type(distribution[key1][key2][key3]) is dict:
                            distribution[key1][key2][key3]["count"] = 0

    # inject counts in the json
    for i in range(0, len(y_classes)):
        pos_class = np.where(y_classes[i] == 1)
        pos_subclass = np.where(y_subclasses[i] == 1)
        pos_leafclass = np.where(y_leafclasses[i] == 1)
        print(
            list_classes[pos_class[0][0]], list_subclasses[pos_subclass[0][0]], list_leaf_classes[pos_leafclass[0][0]]
        )
        if list_classes[pos_class[0][0]] != "no_dataset":
            the_class = list_classes[pos_class[0][0]]
            if list_subclasses[pos_subclass[0][0]] != "nan":
                the_subclass = list_subclasses[pos_subclass[0][0]]
                if list_leaf_classes[pos_leafclass[0][0]] != "nan":
                    the_leafclass = list_leaf_classes[pos_leafclass[0][0]]
                    print(distribution[the_class][the_subclass][the_leafclass])
                    if "count" in distribution[the_class][the_subclass][the_leafclass]:
                        distribution[the_class][the_subclass][the_leafclass]["count"] = (
                            distribution[the_class][the_subclass][the_leafclass]["count"] + 1
                        )
                else:
                    if "count" in distribution[the_class][the_subclass]:
                        distribution[the_class][the_subclass]["count"] = (
                            distribution[the_class][the_subclass]["count"] + 1
                        )
            else:
                if "count" in distribution[the_class]:
                    distribution[the_class]["count"] = distribution[the_class]["count"] + 1

    # save the extended json
    with open("data/textClassification/dataseer/DataTypesWithCounts.json", "w") as outfile:
        json.dump(distribution, outfile, sort_keys=False, indent=4)


if __name__ == "__main__":
    # build_prior_class_distribution()

    parser = argparse.ArgumentParser(
        description="Dataset passage identification and classification for scientific literature based on DeLFT"
    )

    word_embeddings_examples = ["glove-840B", "fasttext-crawl", "word2vec"]
    pretrained_transformers_examples = ["bert-base-cased", "bert-large-cased", "allenai/scibert_scivocab_cased"]

    parser.add_argument("action", help="one of [train, train_eval, classify]")
    parser.add_argument("--fold-count", type=int, default=1)
    parser.add_argument(
        "--architecture", default="gru", help="type of model architecture to be used, one of " + str(architectures)
    )
    parser.add_argument("--cascaded", action="store_true", help="Use models in cascade (train, eval, predict)")
    parser.add_argument(
        "--embedding",
        default=None,
        help="The desired pre-trained word embeddings using their descriptions in the file. "
        + "For local loading, use delft/resources-registry.json. "
        + "Be sure to use here the same name as in the registry, e.g. "
        + str(word_embeddings_examples)
        + " and that the path in the registry to the embedding file is correct on your system.",
    )
    parser.add_argument(
        "--transformer",
        default=None,
        help="The desired pre-trained transformer to be used in the selected architecture. "
        + "For local loading use, delft/resources-registry.json, and be sure to use here the same name as in the registry, e.g. "
        + str(pretrained_transformers_examples)
        + " and that the path in the registry to the model path is correct on your system. "
        + "HuggingFace transformers hub will be used otherwise to fetch the model, see https://huggingface.co/models "
        + "for model names",
    )

    args = parser.parse_args()

    if args.action not in ("train", "train_eval", "classify"):
        print("action not specifed, must be one of [train,train_eval,classify]")

    embeddings_name = args.embedding
    cascaded = args.cascaded
    transformer = args.transformer

    architecture = args.architecture
    if architecture not in architectures:
        print("unknown model architecture, must be one of " + str(architectures))

    if transformer is None and embeddings_name is None:
        # default word embeddings
        embeddings_name = "glove-840B"

    if args.action == "train":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")

        train(
            embeddings_name=embeddings_name,
            fold_count=args.fold_count,
            architecture=architecture,
            transformer=transformer,
            cascaded=cascaded,
        )

    if args.action == "train_eval":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")
        y_test = train_and_eval(
            embeddings_name=embeddings_name,
            fold_count=args.fold_count,
            architecture=architecture,
            transformer=transformer,
            cascaded=cascaded,
        )

    if args.action == "classify":
        someTexts = [
            "Labeling yield and radiochemical purity was analyzed by instant thin layered chromatography (ITLC).",
            "NOESY and ROESY spectra 64,65 were collected with typically 128 scans per t1 increment, with the residual water signal removed by the WATERGATE sequence and 1 s relaxation time.",
            "The concentrations of Cd and Pb in feathers were measured by furnace atomic absorption spectrometry (VARIAN 240Z).",
        ]
        result = classify(someTexts, "json", architecture=architecture)
        print(json.dumps(result, sort_keys=False, indent=4, ensure_ascii=False))



from typing import Dict, List, Tuple

__author__ = "@de-code"

"""
Utility class from:
https://github.com/elifesciences/sciencebeam-trainer-delft/blob/develop/sciencebeam_trainer_delft/utils/misc.py
"""


def parse_number_range(expr: str) -> List[int]:
    fragments = expr.split("-")
    if len(fragments) == 1:
        return [int(expr)]
    if len(fragments) == 2:
        return list(range(int(fragments[0]), int(fragments[1]) + 1))
    raise ValueError("invalid number range: %s" % fragments)


def parse_number_ranges(expr: str) -> List[int]:
    if not expr:
        return []
    numbers = []
    for fragment in expr.split(","):
        numbers.extend(parse_number_range(fragment))
    return numbers


def parse_key_value(expr: str) -> Tuple[str, str]:
    key, value = expr.split("=", maxsplit=1)
    return key.strip(), value.strip()


def parse_dict(expr: str, delimiter: str = "|") -> Dict[str, str]:
    if not expr:
        return {}
    d = {}
    for fragment in expr.split(delimiter):
        key, value = parse_key_value(fragment)
        d[key] = value
    return d


def merge_dicts(dict_list: List[dict]) -> dict:
    result = {}
    for d in dict_list:
        result.update(d)
    return result


def print_parameters(model_config, training_config):
    print("---")
    print("max_epoch:", training_config.max_epoch)
    print("early_stop:", training_config.early_stop)
    if training_config.early_stop:
        print("patience:", training_config.patience)
    print("batch_size (training):", model_config.batch_size)

    if hasattr(model_config, "max_sequence_length"):
        print("max_sequence_length:", model_config.max_sequence_length)

    if hasattr(model_config, "maxlen"):
        print("maxlen:", model_config.maxlen)

    print("model_name:", model_config.model_name)
    print("learning_rate: ", training_config.learning_rate)

    if hasattr(training_config, "num_workers"):
        print("num_workers: ", training_config.num_workers)

    if (
        hasattr(training_config, "class_weights")
        and training_config.class_weights is not None
        and hasattr(model_config, "list_classes")
    ):
        weight_summary = ""
        for indx, class_name in enumerate(model_config.list_classes):
            if indx != 0:
                weight_summary += ", "
            weight_summary += class_name + ": " + str(training_config.class_weights[indx])
        print("class_weights:", weight_summary)

    print("---")



import pickle

import numpy as np
import pytest

from delft.utilities.Embeddings import _check_lmdb_format


class TestCheckLmdbFormat:
    def test_accepts_raw_float32(self):
        """Raw float32 bytes should pass validation without error."""
        vector = np.random.rand(300).astype(np.float32)
        raw_bytes = vector.tobytes()
        _check_lmdb_format(raw_bytes)  # should not raise

    def test_rejects_pickle(self):
        """Pickled numpy arrays should raise ValueError."""
        vector = np.zeros(300, dtype=np.float32)
        pickled = pickle.dumps(vector)
        with pytest.raises(ValueError, match="legacy pickle format"):
            _check_lmdb_format(pickled)

    def test_accepts_short_values(self):
        """Values shorter than 2 bytes should pass (edge case)."""
        _check_lmdb_format(b"")
        _check_lmdb_format(b"\x00")

    def test_no_false_positive_on_0x80(self):
        """A float32 vector that happens to start with 0x80 should not be rejected."""
        # Craft a float32 array whose first byte is 0x80 (e.g., -0.0 in IEEE 754 is 0x80000000)
        vector = np.array([-0.0] + [1.0] * 299, dtype=np.float32)
        raw_bytes = vector.tobytes()
        assert raw_bytes[0] == 0x00  # -0.0 little-endian is 00 00 00 80

        # Directly craft bytes starting with 0x80 to be sure
        crafted = bytes([0x80, 0x03]) + b"\x00" * 1198  # 0x80 + protocol 3, but no 'numpy'
        _check_lmdb_format(crafted)  # should not raise (no b'numpy' in first 50 bytes)

    def test_rejects_all_pickle_protocols(self):
        """Pickle protocols 2-5 should all be detected."""
        vector = np.zeros(300, dtype=np.float32)
        for protocol in (2, 3, 4, 5):
            pickled = pickle.dumps(vector, protocol=protocol)
            with pytest.raises(ValueError, match="legacy pickle format"):
                _check_lmdb_format(pickled)



from __future__ import absolute_import

"""
Originally from Philipp Gross, https://github.com/phipleg/keras/blob/crf/keras/layers/crf.py

Adapted to Keras/tensorflow 2 by your DeLFT servitor.

Note: in this version, zero masking is not working with TF2, so do not use mask_zero=True in
your architecture when using this CRF layer

Note: relying on tensorflow.compat.v1.keras function usages, it's probably ok.
"""


import tensorflow as tf
from tf_keras import backend as K
from tf_keras import constraints, initializers, regularizers
from tf_keras.layers import InputSpec, Layer


def path_energy(y, x, U, b_start=None, b_end=None, mask=None):
    """Calculates the energy of a tag path y for a given input x (with mask),
    transition energies U and boundary energies b_start, b_end."""
    x = add_boundary_energy(x, b_start, b_end, mask)
    return path_energy0(y, x, U, mask)


def path_energy0(y, x, U, mask=None):
    """Path energy without boundary potential handling."""
    n_classes = K.shape(x)[2]
    y_one_hot = K.one_hot(y, n_classes)

    # Tag path energy
    energy = K.sum(x * y_one_hot, 2)
    energy = K.sum(energy, 1)

    # Transition energy
    y_t = y[:, :-1]
    y_tp1 = y[:, 1:]
    U_flat = K.reshape(U, [-1])
    # Convert 2-dim indices (y_t, y_tp1) of U to 1-dim indices of U_flat:
    flat_indices = y_t * n_classes + y_tp1
    U_y_t_tp1 = K.gather(U_flat, flat_indices)

    if mask is not None:
        mask = K.cast(mask, K.floatx())
        y_t_mask = mask[:, :-1]
        y_tp1_mask = mask[:, 1:]
        U_y_t_tp1 *= y_t_mask * y_tp1_mask

    energy += K.sum(U_y_t_tp1, axis=1)

    return energy


def sparse_chain_crf_loss(y, x, U, b_start=None, b_end=None, mask=None):
    """Given the true sparsely encoded tag sequence y, input x (with mask),
    transition energies U, boundary energies b_start and b_end, it computes
    the loss function of a Linear Chain Conditional Random Field:
    loss(y, x) = NNL(P(y|x)), where P(y|x) = exp(E(y, x)) / Z.
    So, loss(y, x) = - E(y, x) + log(Z)
    Here, E(y, x) is the tag path energy, and Z is the normalization constant.
    The values log(Z) is also called free energy.
    """
    x = add_boundary_energy(x, b_start, b_end, mask)
    energy = path_energy0(y, x, U, mask)
    energy -= free_energy0(x, U, mask)
    return K.expand_dims(-energy, -1)


def chain_crf_loss(y, x, U, b_start=None, b_end=None, mask=None):
    """Variant of sparse_chain_crf_loss but with one-hot encoded tags y."""
    y_sparse = K.argmax(y, -1)
    y_sparse = K.cast(y_sparse, "int32")
    return sparse_chain_crf_loss(y_sparse, x, U, b_start, b_end, mask)


def add_boundary_energy(x, b_start=None, b_end=None, mask=None):
    """Given the observations x, it adds the start boundary energy b_start (resp.
    end boundary energy b_end on the start (resp. end) elements and multiplies
    the mask."""
    if mask is None:
        if b_start is not None:
            x = K.concatenate([x[:, :1, :] + b_start, x[:, 1:, :]], axis=1)
        if b_end is not None:
            x = K.concatenate([x[:, :-1, :], x[:, -1:, :] + b_end], axis=1)
    else:
        mask = K.cast(mask, K.floatx())
        mask = K.expand_dims(mask, 2)
        x *= mask
        if b_start is not None:
            mask_r = K.concatenate([K.zeros_like(mask[:, :1]), mask[:, :-1]], axis=1)
            start_mask = K.cast(K.greater(mask, mask_r), K.floatx())
            x = x + start_mask * b_start
        if b_end is not None:
            mask_l = K.concatenate([mask[:, 1:], K.zeros_like(mask[:, -1:])], axis=1)
            end_mask = K.cast(K.greater(mask, mask_l), K.floatx())
            x = x + end_mask * b_end
    return x


def viterbi_decode(x, U, b_start=None, b_end=None, mask=None):
    """Computes the best tag sequence y for a given input x, i.e. the one that
    maximizes the value of path_energy."""
    x = add_boundary_energy(x, b_start, b_end, mask)

    alpha_0 = x[:, 0, :]
    gamma_0 = K.zeros_like(alpha_0)
    initial_states = [gamma_0, alpha_0]
    _, gamma = _forward(
        x, lambda B: [K.cast(K.argmax(B, axis=1), K.floatx()), K.max(B, axis=1)], initial_states, U, mask
    )
    y = _backward(gamma, mask)
    return y


def free_energy(x, U, b_start=None, b_end=None, mask=None):
    """Computes efficiently the sum of all path energies for input x, when
    runs over all possible tag sequences."""
    x = add_boundary_energy(x, b_start, b_end, mask)
    return free_energy0(x, U, mask)


def free_energy0(x, U, mask=None):
    """Free energy without boundary potential handling."""
    initial_states = [x[:, 0, :]]
    last_alpha, _ = _forward(x, lambda B: [tf.reduce_logsumexp(B, axis=1)], initial_states, U, mask)
    return last_alpha[:, 0]


def _forward(x, reduce_step, initial_states, U, mask=None):
    """Forward recurrence of the linear chain crf."""

    def _forward_step(energy_matrix_t, states):
        alpha_tm1 = states[-1]
        new_states = reduce_step(K.expand_dims(alpha_tm1, 2) + energy_matrix_t)
        return new_states[0], new_states

    U_shared = K.expand_dims(K.expand_dims(U, 0), 0)

    if mask is not None:
        mask = K.cast(mask, K.floatx())
        mask_U = K.expand_dims(K.expand_dims(mask[:, :-1] * mask[:, 1:], 2), 3)
        U_shared = U_shared * mask_U

    inputs = K.expand_dims(x[:, 1:, :], 2) + U_shared
    inputs = K.concatenate([inputs, K.zeros_like(inputs[:, -1:, :, :])], axis=1)

    last, values, _ = K.rnn(_forward_step, inputs, initial_states)
    return last, values


def batch_gather(reference, indices):
    ref_shape = K.shape(reference)
    batch_size = ref_shape[0]
    n_classes = ref_shape[1]
    flat_indices = K.arange(0, batch_size) * n_classes + K.flatten(indices)
    return K.gather(K.flatten(reference), flat_indices)


def _backward(gamma, mask):
    """Backward recurrence of the linear chain crf."""
    gamma = K.cast(gamma, "int32")

    def _backward_step(gamma_t, states):
        y_tm1 = K.squeeze(states[0], 0)
        y_t = batch_gather(gamma_t, y_tm1)
        return y_t, [K.expand_dims(y_t, 0)]

    initial_states = [K.expand_dims(K.zeros_like(gamma[:, 0, 0]), 0)]
    _, y_rev, _ = K.rnn(_backward_step, gamma, initial_states, go_backwards=True)
    y = K.reverse(y_rev, 1)

    if mask is not None:
        mask = K.cast(mask, dtype="int32")
        # mask output
        y *= mask
        # set masked values to -1
        y += -(1 - mask)
    return y


class ChainCRF(Layer):
    """A Linear Chain Conditional Random Field output layer.
    It carries the loss function and its weights for computing
    the global tag sequence scores. While training it acts as
    the identity function that passes the inputs to the subsequently
    used loss function. While testing it applies Viterbi decoding
    and returns the best scoring tag sequence as one-hot encoded vectors.
    # Arguments
        init: weight initialization function for chain energies U.
            Can be the name of an existing function (str),
            or a Theano function (see: [initializers](../initializers.md)).
        U_regularizer: instance of [WeightRegularizer](../regularizers.md)
            (eg. L1 or L2 regularization), applied to the transition weight matrix.
        b_start_regularizer: instance of [WeightRegularizer](../regularizers.md),
            applied to the start bias b.
        b_end_regularizer: instance of [WeightRegularizer](../regularizers.md)
            module, applied to the end bias b.
        b_start_constraint: instance of the [constraints](../constraints.md)
            module, applied to the start bias b.
        b_end_constraint: instance of the [constraints](../constraints.md)
            module, applied to the end bias b.
        weights: list of Numpy arrays for initializing [U, b_start, b_end].
            Thus it should be a list of 3 elements of shape
            [(n_classes, n_classes), (n_classes, ), (n_classes, )]
    # Input shape
        3D tensor with shape `(nb_samples, timesteps, nb_classes)`, where
        ´timesteps >= 2`and `nb_classes >= 2`.
    # Output shape
        Same shape as input.
    # Masking
        This layer supports masking for input sequences of variable length.
    # Example
    ```python
    # As the last layer of sequential layer with
    # model.output_shape == (None, timesteps, nb_classes)
    crf = ChainCRF()
    model.add(crf)
    # now: model.output_shape == (None, timesteps, nb_classes)
    # Compile model with chain crf loss (and one-hot encoded labels) and accuracy
    model.compile(loss=crf.loss, optimizer='sgd', metrics=['accuracy'])
    # Alternatively, compile model with sparsely encoded labels and sparse accuracy:
    model.compile(loss=crf.sparse_loss, optimizer='sgd', metrics=['sparse_categorical_accuracy'])
    ```
    # Gotchas
    ## Model loading
    When you want to load a saved model that has a crf output, then loading
    the model with 'keras.models.load_model' won't work properly because
    the reference of the loss function to the transition parameters is lost. To
    fix this, you need to use the parameter 'custom_objects' as follows:
    ```python
    from keras.layer.crf import create_custom_objects:
    model = keras.models.load_model(filename, custom_objects=create_custom_objects())
    ```
    ## Temporal sample weights
    Given a ChainCRF instance crf both loss functions, crf.loss and crf.sparse_loss
    return a tensor of shape (batch_size, 1) and not (batch_size, maxlen).
    that sample weighting in temporal mode.
    """

    def __init__(
        self,
        init="glorot_uniform",
        U_regularizer=None,
        b_start_regularizer=None,
        b_end_regularizer=None,
        U_constraint=None,
        b_start_constraint=None,
        b_end_constraint=None,
        weights=None,
        **kwargs,
    ):
        super(ChainCRF, self).__init__(**kwargs)
        self.init = initializers.get(init)
        self.U_regularizer = regularizers.get(U_regularizer)
        self.b_start_regularizer = regularizers.get(b_start_regularizer)
        self.b_end_regularizer = regularizers.get(b_end_regularizer)
        self.U_constraint = constraints.get(U_constraint)
        self.b_start_constraint = constraints.get(b_start_constraint)
        self.b_end_constraint = constraints.get(b_end_constraint)

        self.initial_weights = weights

        self.supports_masking = True
        self.uses_learning_phase = True
        self.input_spec = [InputSpec(ndim=3)]

    def compute_output_shape(self, input_shape):
        assert input_shape and len(input_shape) == 3
        return (input_shape[0], input_shape[1], input_shape[2])

    def compute_mask(self, input, mask=None):
        if mask is not None:
            return K.any(mask, axis=1)
        return mask

    def _fetch_mask(self):
        mask = None
        if self._inbound_nodes:
            # mask = self._inbound_nodes[0].input_masks[0]
            mask = self.get_input_mask_at(0)
        return mask

    def build(self, input_shape):
        assert len(input_shape) == 3
        n_classes = input_shape[2]
        n_steps = input_shape[1]
        assert n_steps is None or n_steps >= 2
        self.input_spec = [InputSpec(dtype=K.floatx(), shape=(None, n_steps, n_classes))]

        self.U = self.add_weight(
            shape=(n_classes, n_classes),
            initializer=self.init,
            name="U",
            regularizer=self.U_regularizer,
            constraint=self.U_constraint,
        )

        self.b_start = self.add_weight(
            shape=(n_classes,),
            initializer="zero",
            name="b_start",
            regularizer=self.b_start_regularizer,
            constraint=self.b_start_constraint,
        )

        self.b_end = self.add_weight(
            shape=(n_classes,),
            initializer="zero",
            name="b_end",
            regularizer=self.b_end_regularizer,
            constraint=self.b_end_constraint,
        )

        if self.initial_weights is not None:
            self.set_weights(self.initial_weights)
            del self.initial_weights

        self.built = True

    def call(self, x, mask=None):
        y_pred = viterbi_decode(x, self.U, self.b_start, self.b_end, mask)
        nb_classes = self.input_spec[0].shape[2]
        y_pred_one_hot = K.one_hot(y_pred, nb_classes)
        return K.in_train_phase(x, y_pred_one_hot)

    def loss(self, y_true, y_pred):
        """Linear Chain Conditional Random Field loss function."""
        mask = self._fetch_mask()
        return chain_crf_loss(y_true, y_pred, self.U, self.b_start, self.b_end, mask)

    def sparse_crf_loss_masked(self, y_true, y_pred):
        mask_value = 0
        y_true_masked = tf.ragged.boolean_mask(y_true, tf.not_equal(y_true, mask_value)).to_tensor()
        y_pred_masked = tf.ragged.boolean_mask(y_pred, tf.not_equal(y_true, mask_value)).to_tensor()
        return self.sparse_loss(y_true_masked, y_pred_masked)

    def sparse_crf_loss_bert_masked(self, y_true, y_pred):
        mask_value = 0

        # create a boolean mask with False for labels to be ignored
        special_mask = tf.not_equal(y_true, mask_value)
        special_mask = tf.cast(special_mask, tf.float32)
        # apply the mask to true & prediction vectors, it will put to 0
        # weights for label to ignore, normally neutralizing them for
        # loss calculation
        y_pred_masked = tf.multiply(y_pred, tf.expand_dims(special_mask, -1))

        return self.sparse_loss(y_true, y_pred_masked)

    def sparse_loss(self, y_true, y_pred):
        """
        Linear Chain Conditional Random Field loss function with sparse
        tag sequences.
        """
        y_true = tf.cast(y_true, "int32")

        # note: nothing to squeeze here with sparse labels which are 2D
        # y_true = tf.squeeze(y_true, [2])
        mask = self._fetch_mask()
        return sparse_chain_crf_loss(y_true, y_pred, self.U, self.b_start, self.b_end, mask)

    def get_config(self):
        config = {
            "init": initializers.serialize(self.init),
            "U_regularizer": regularizers.serialize(self.U_regularizer),
            "b_start_regularizer": regularizers.serialize(self.b_start_regularizer),
            "b_end_regularizer": regularizers.serialize(self.b_end_regularizer),
            "U_constraint": constraints.serialize(self.U_constraint),
            "b_start_constraint": constraints.serialize(self.b_start_constraint),
            "b_end_constraint": constraints.serialize(self.b_end_constraint),
        }
        base_config = super(ChainCRF, self).get_config()
        return dict(list(base_config.items()) + list(config.items()))


def create_custom_objects():
    """Returns the custom objects, needed for loading a persisted model."""
    instanceHolder = {"instance": None}

    class ClassWrapper(ChainCRF):
        def __init__(self, *args, **kwargs):
            instanceHolder["instance"] = self
            super(ClassWrapper, self).__init__(*args, **kwargs)

    def loss(*args):
        method = getattr(instanceHolder["instance"], "loss")
        return method(*args)

    def sparse_loss(*args):
        method = getattr(instanceHolder["instance"], "sparse_loss")
        return method(*args)

    return {"ChainCRF": ClassWrapper, "loss": loss, "sparse_loss": sparse_loss}



from delft.utilities.Utilities import truecase_sentence

"""
Convert CoNLL tokens casing via truecase
"""


def convert(input_file, output_file):
    with open(output_file, "w") as outfile:
        with open(input_file, "r+", encoding="UTF-8") as infile:
            words, tags = [], []
            for original_line in infile:
                line = original_line.rstrip()
                if len(line) == 0 or line.startswith("-DOCSTART-") or line.startswith("#begin document"):
                    if len(words) != 0:
                        words = truecase_sentence(words)
                        for i in range(len(words)):
                            outfile.write(words[i])
                            outfile.write("\t")
                            outfile.write(tags[i])
                            outfile.write("\n")
                        words, tags = [], []
                    outfile.write(original_line)
                else:
                    if len(line.split("\t")) == 2:
                        word, tag = line.split("\t")
                    else:
                        word, _, tag = line.split("\t")
                    words.append(word)
                    tags.append(tag)


if __name__ == "__main__":
    convert("../../data/sequenceLabelling/CoNLL-2003/eng.testa", "../../data/sequenceLabelling/CoNLL-2003-TC/eng.testa")
    convert("../../data/sequenceLabelling/CoNLL-2003/eng.testb", "../../data/sequenceLabelling/CoNLL-2003-TC/eng.testb")
    convert("../../data/sequenceLabelling/CoNLL-2003/eng.train", "../../data/sequenceLabelling/CoNLL-2003-TC/eng.train")



# -*- coding: utf-8 -*-

import argparse
import json
import os
import re
import time
from datetime import datetime

from sklearn.model_selection import train_test_split

from delft.sequenceLabelling import Sequence
from delft.sequenceLabelling.reader import load_data_and_labels_crf_file
from delft.utilities.Utilities import longest_row, t_or_f

MODEL_LIST = [
    "affiliation-address",
    "citation",
    "date",
    "header",
    "name-citation",
    "name-header",
    "software",
    "figure",
    "table",
    "reference-segmenter",
    "segmentation",
    "funding-acknowledgement",
    "patent-citation",
]


def find_latest_train_file(model: str) -> str:
    """
    Find the latest training file for the given model based on the date in the filename.

    Files are expected to have names like: {model}-YYMMDD.train
    Returns the path to the latest file, or None if no files found.
    """
    data_dir = f"data/sequenceLabelling/grobid/{model}"
    if not os.path.exists(data_dir):
        return None

    # Pattern: model-YYMMDD.train
    pattern = re.compile(rf"^{re.escape(model)}-(\d{{6}})\.train$")

    latest_file = None
    latest_date = None

    for filename in os.listdir(data_dir):
        match = pattern.match(filename)
        if match:
            date_str = match.group(1)
            try:
                # Parse YYMMDD format
                date = datetime.strptime(date_str, "%y%m%d")
                if latest_date is None or date > latest_date:
                    latest_date = date
                    latest_file = os.path.join(data_dir, filename)
            except ValueError:
                continue

    return latest_file


def configure(
    model,
    architecture,
    output_path=None,
    max_sequence_length=-1,
    batch_size=-1,
    embeddings_name=None,
    max_epoch=-1,
    patience=-1,
    early_stop=None,
):
    """
    Set up the default parameters based on the model type.
    """
    if output_path:
        model_name = model
    else:
        model_name = "grobid-" + model

    multiprocessing = True
    o_early_stop = True

    if architecture and "BERT" in architecture:
        # architectures with some transformer layer/embeddings inside

        # non-default settings per model
        if model == "citation":
            if max_sequence_length == -1:
                max_sequence_length = 200
            if batch_size == -1:
                batch_size = 20
        elif model == "header":
            if max_sequence_length == -1:
                max_sequence_length = 512
            if batch_size == -1:
                batch_size = 6
        elif model == "date":
            if max_sequence_length == -1:
                max_sequence_length = 30
            if batch_size == -1:
                batch_size = 80
        elif model == "affiliation-address":
            if max_sequence_length == -1:
                max_sequence_length = 200
            if batch_size == -1:
                batch_size = 20
        elif model.startswith("software"):
            # class are more unbalanced, so we need to extend the batch size as much as we can
            if batch_size == -1:
                batch_size = 8
            if max_sequence_length == -1:
                max_sequence_length = 512
            o_early_stop = False
            if max_epoch == -1:
                max_epoch = 30
        elif model.startswith("funding-acknowledgement"):
            if max_sequence_length == -1:
                max_sequence_length = 512
            if batch_size == -1:
                batch_size = 8
        elif model == "patent-citation":
            if max_sequence_length == -1:
                max_sequence_length = 512
            if batch_size == -1:
                batch_size = 8

        # default when no value provided by command line or model-specific
        if batch_size == -1:
            # default
            batch_size = 20
        if max_sequence_length == -1:
            # default
            max_sequence_length = 200

        if max_sequence_length > 512:
            # 512 is the largest sequence for BERT input
            max_sequence_length = 512

        embeddings_name = None
    else:
        # RNN-only architectures
        if model == "citation":
            if max_sequence_length == -1:
                max_sequence_length = 500
            if batch_size == -1:
                batch_size = 30
        elif model == "header":
            max_epoch = 80
            if max_sequence_length == -1:
                max_sequence_length = 2500
            if batch_size == -1:
                batch_size = 9
        elif model == "date":
            if max_sequence_length == -1:
                max_sequence_length = 50
            if batch_size == -1:
                batch_size = 60
        elif model == "affiliation-address":
            if max_sequence_length == -1:
                max_sequence_length = 600
            if batch_size == -1:
                batch_size = 20
        elif model.startswith("software"):
            if batch_size == -1:
                batch_size = 20
            if max_sequence_length == -1:
                max_sequence_length = 1500
            multiprocessing = False
        elif model == "reference-segmenter":
            if batch_size == -1:
                batch_size = 5
            if max_sequence_length == -1:
                max_sequence_length = 3000
        elif model == "funding-acknowledgement":
            if batch_size == -1:
                batch_size = 30
            if max_sequence_length == -1:
                max_sequence_length = 800
        elif model == "patent-citation":
            if batch_size == -1:
                batch_size = 40
            if max_sequence_length == -1:
                max_sequence_length = 1000

    model_name += "-" + architecture

    if batch_size == -1:
        batch_size = 20

    if max_sequence_length == -1:
        max_sequence_length = 3000

    if max_epoch == -1:
        max_epoch = 60

    if patience == -1:
        patience = 5

    if early_stop is not None:
        o_early_stop = early_stop

    return (
        batch_size,
        max_sequence_length,
        model_name,
        embeddings_name,
        max_epoch,
        multiprocessing,
        o_early_stop,
        patience,
    )


# train a GROBID model with all available data


def train(
    model,
    embeddings_name=None,
    architecture=None,
    transformer=None,
    input_path=None,
    output_path=None,
    features_indices=None,
    max_sequence_length=-1,
    batch_size=-1,
    max_epoch=-1,
    incremental=False,
    input_model_path=None,
    patience=-1,
    learning_rate=None,
    early_stop=None,
    multi_gpu=False,
    report_to_wandb=False,
    num_workers=1,
):

    print("Loading data...")
    if input_path is None:
        input_path = find_latest_train_file(model)
        if input_path is None:
            raise ValueError(f"No training file found for model '{model}' in data/sequenceLabelling/grobid/{model}/")
        print(f"Using latest training file: {input_path}")
    x_all, y_all, f_all = load_data_and_labels_crf_file(input_path)

    print(len(x_all), "total sequences")

    x_train, x_valid, y_train, y_valid, f_train, f_valid = train_test_split(
        x_all, y_all, f_all, test_size=0.1, shuffle=True
    )

    print(len(x_train), "train sequences")
    print(len(x_valid), "validation sequences")

    print("\nmax train sequence length:", str(longest_row(x_train)))
    print("max validation sequence length:", str(longest_row(x_valid)))

    (batch_size, max_sequence_length, model_name, embeddings_name, max_epoch, multiprocessing, early_stop, patience) = (
        configure(
            model,
            architecture,
            output_path,
            max_sequence_length,
            batch_size,
            embeddings_name,
            max_epoch,
            patience,
            early_stop,
        )
    )

    model = Sequence(
        model_name,
        recurrent_dropout=0.50,
        embeddings_name=embeddings_name,
        architecture=architecture,
        transformer_name=transformer,
        batch_size=batch_size,
        max_sequence_length=max_sequence_length,
        features_indices=features_indices,
        max_epoch=max_epoch,
        multiprocessing=multiprocessing,
        num_workers=num_workers,
        early_stop=early_stop,
        patience=patience,
        learning_rate=learning_rate,
        report_to_wandb=report_to_wandb,
    )

    if incremental:
        if input_model_path is not None:
            model.load(input_model_path)
        elif output_path is not None:
            model.load(output_path)
        else:
            model.load()

    start_time = time.time()
    model.train(x_train, y_train, f_train, x_valid, y_valid, f_valid, incremental=incremental, multi_gpu=multi_gpu)

    runtime = round(time.time() - start_time, 3)
    print("training runtime: %s seconds " % (runtime))

    # saving the model
    if output_path:
        model.save(output_path)
    else:
        model.save()


# split data, train a GROBID model and evaluate it
def train_eval(
    model,
    embeddings_name=None,
    architecture="BidLSTM_CRF",
    transformer=None,
    input_path=None,
    output_path=None,
    fold_count=1,
    features_indices=None,
    max_sequence_length=-1,
    batch_size=-1,
    max_epoch=-1,
    incremental=False,
    input_model_path=None,
    patience=-1,
    learning_rate=None,
    early_stop=None,
    multi_gpu=False,
    report_to_wandb=False,
):

    print("Loading data...")
    if input_path is None:
        input_path = find_latest_train_file(model)
        if input_path is None:
            raise ValueError(f"No training file found for model '{model}' in data/sequenceLabelling/grobid/{model}/")
        print(f"Using latest training file: {input_path}")
    x_all, y_all, f_all = load_data_and_labels_crf_file(input_path)

    x_train_all, x_eval, y_train_all, y_eval, f_train_all, f_eval = train_test_split(
        x_all, y_all, f_all, test_size=0.1, shuffle=True
    )
    x_train, x_valid, y_train, y_valid, f_train, f_valid = train_test_split(
        x_train_all, y_train_all, f_train_all, test_size=0.1
    )

    print(len(x_train), "train sequences")
    print(len(x_valid), "validation sequences")
    print(len(x_eval), "evaluation sequences")

    print("\nmax train sequence length:", str(longest_row(x_train)))
    print("max validation sequence length:", str(longest_row(x_valid)))
    print("max evaluation sequence length:", str(longest_row(x_eval)))

    batch_size, max_sequence_length, model_name, embeddings_name, max_epoch, multiprocessing, early_stop, patience = (
        configure(
            model,
            architecture,
            output_path,
            max_sequence_length,
            batch_size,
            embeddings_name,
            max_epoch,
            patience,
            early_stop,
        )
    )

    model = Sequence(
        model_name,
        architecture=architecture,
        embeddings_name=embeddings_name,
        max_sequence_length=max_sequence_length,
        recurrent_dropout=0.50,
        batch_size=batch_size,
        learning_rate=learning_rate,
        max_epoch=max_epoch,
        early_stop=early_stop,
        patience=patience,
        fold_number=fold_count,
        multiprocessing=multiprocessing,
        features_indices=features_indices,
        transformer_name=transformer,
        report_to_wandb=report_to_wandb,
    )

    if incremental:
        if input_model_path is not None:
            model.load(input_model_path)
        elif output_path is not None:
            model.load(output_path)
        else:
            model.load()

    start_time = time.time()

    if fold_count == 1:
        model.train(
            x_train,
            y_train,
            f_train=f_train,
            x_valid=x_valid,
            y_valid=y_valid,
            f_valid=f_valid,
            incremental=incremental,
            multi_gpu=multi_gpu,
        )
    else:
        model.train_nfold(
            x_train,
            y_train,
            f_train=f_train,
            x_valid=x_valid,
            y_valid=y_valid,
            f_valid=f_valid,
            incremental=incremental,
            multi_gpu=multi_gpu,
        )

    runtime = round(time.time() - start_time, 3)
    print("training runtime: %s seconds " % runtime)

    # evaluation
    print("\nEvaluation:")
    model.eval(x_eval, y_eval, features=f_eval)

    # saving the model (must be called after eval for multiple fold training)
    if output_path:
        model.save(output_path)
    else:
        model.save()


# split data, train a GROBID model and evaluate it
def eval_(model, input_path=None, architecture="BidLSTM_CRF", report_to_wandb=False):
    print("Loading data...")
    if input_path is None:
        # it should never be the case
        print(
            "A Grobid evaluation data file must be specified for evaluating a grobid model for the eval action, use parameter --input "
        )
        return
    else:
        x_all, y_all, f_all = load_data_and_labels_crf_file(input_path)

    print(len(x_all), "evaluation sequences")

    print("max evaluation sequence length:", str(longest_row(x_all)))

    model_name = "grobid-" + model
    model_name += "-" + architecture

    start_time = time.time()

    # load the model
    model = Sequence(model_name, report_to_wandb=report_to_wandb)
    model.load()

    # evaluation
    print("\nEvaluation:")
    model.eval(x_all, y_all, features=f_all)

    runtime = round(time.time() - start_time, 3)
    print("Evaluation runtime: %s seconds " % (runtime))


# annotate a list of texts, this is relevant only of models taking only text as input
# (so not text with layout information)
def annotate_text(texts, model, output_format, architecture="BidLSTM_CRF", features=None, multi_gpu=False):
    annotations = []

    # load model
    model_name = "grobid-" + model
    model_name += "-" + architecture

    model = Sequence(model_name)
    model.load()

    start_time = time.time()

    annotations = model.tag(texts, output_format, features=features, multi_gpu=multi_gpu)
    runtime = round(time.time() - start_time, 3)

    if output_format == "json":
        annotations["runtime"] = runtime
    else:
        print("runtime: %s seconds " % (runtime))

    return annotations


class Tasks:
    TRAIN = "train"
    TRAIN_EVAL = "train_eval"
    EVAL = "eval"
    TAG = "tag"


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Trainer for GROBID models using the DeLFT library")

    actions = [Tasks.TRAIN, Tasks.TRAIN_EVAL, Tasks.EVAL, Tasks.TAG]

    architectures_word_embeddings = [
        "BidLSTM",
        "BidLSTM_CRF",
        "BidLSTM_ChainCRF",
        "BidLSTM_CNN_CRF",
        "BidLSTM_CNN_CRF",
        "BidGRU_CRF",
        "BidLSTM_CNN",
        "BidLSTM_CRF_CASING",
        "BidLSTM_CRF_FEATURES",
        "BidLSTM_ChainCRF_FEATURES",
    ]

    word_embeddings_examples = ["glove-840B", "fasttext-crawl", "word2vec"]

    architectures_transformers_based = [
        "BERT",
        "BERT_FEATURES",
        "BERT_CRF",
        "BERT_ChainCRF",
        "BERT_CRF_FEATURES",
        "BERT_ChainCRF_FEATURES",
        "BERT_CRF_CHAR",
        "BERT_CRF_CHAR_FEATURES",
    ]

    architectures = architectures_word_embeddings + architectures_transformers_based

    pretrained_transformers_examples = ["bert-base-cased", "bert-large-cased", "allenai/scibert_scivocab_cased"]

    parser.add_argument("model", help="Name of the model.")
    parser.add_argument("action", choices=actions)
    parser.add_argument(
        "--fold-count", type=int, default=1, help="Number of fold to use when evaluating with n-fold cross validation."
    )
    parser.add_argument("--architecture", help="Type of model architecture to be used, one of " + str(architectures))
    parser.add_argument("--output", help="Directory where to save a trained model.")

    parser.add_argument(
        "--embedding",
        default=None,
        help="The desired pre-trained word embeddings using their descriptions in the file. "
        + "For local loading, use delft/resources-registry.json. "
        + "Be sure to use here the same name as in the registry, e.g. "
        + str(word_embeddings_examples)
        + " and that the path in the registry to the embedding file is correct on your system.",
    )
    parser.add_argument(
        "--transformer",
        default=None,
        help="The desired pre-trained transformer to be used in the selected architecture. "
        + "For local loading use, delft/resources-registry.json, and be sure to use here the same name as in the registry, e.g. "
        + str(pretrained_transformers_examples)
        + " and that the path in the registry to the model path is correct on your system. "
        + "HuggingFace transformers hub will be used otherwise to fetch the model, see https://huggingface.co/models "
        + "for model names",
    )

    parser.add_argument(
        "--input",
        help="Grobid data file to be used for training (train action), for training and "
        + "evaluation (train_eval action) or just for evaluation (eval action).",
    )
    parser.add_argument(
        "--incremental", action="store_true", help="training is incremental, starting from existing model if present"
    )
    parser.add_argument(
        "--input-model",
        help="In case of incremental training, path to an existing model to be used "
        + "to start the training, instead of the default one.",
    )
    parser.add_argument("--max-sequence-length", type=int, default=-1, help="max-sequence-length parameter to be used.")
    parser.add_argument("--batch-size", type=int, default=-1, help="batch-size parameter to be used.")
    parser.add_argument(
        "--patience",
        type=int,
        default=-1,
        help="patience, number of extra epochs to perform after the best epoch before stopping a training.",
    )
    parser.add_argument("--learning-rate", type=float, default=None, help="Initial learning rate")

    parser.add_argument("--max-epoch", type=int, default=-1, help="Maximum number of epochs for training.")
    parser.add_argument(
        "--early-stop",
        type=t_or_f,
        default=None,
        help="Force early training termination when metrics scores are not improving "
        + "after a number of epochs equals to the patience parameter.",
    )

    parser.add_argument(
        "--multi-gpu",
        default=False,
        help="Enable the support for distributed computing (the batch size needs to be set accordingly using --batch-size)",
        action="store_true",
    )

    parser.add_argument(
        "--num-workers",
        type=int,
        default=1,
        help="Number of worker processes for data loading (default: 1, use 0 or 1 for no multiprocessing)",
    )

    parser.add_argument(
        "--wandb",
        default=False,
        help="Enable the logging of the training using Weights and Biases",
        action="store_true",
    )

    args = parser.parse_args()

    model = args.model
    action = args.action
    architecture = args.architecture
    output = args.output
    input_path = args.input
    input_model_path = args.input_model
    embeddings_name = args.embedding
    max_sequence_length = args.max_sequence_length
    batch_size = args.batch_size
    transformer = args.transformer
    incremental = args.incremental
    patience = args.patience
    learning_rate = args.learning_rate
    max_epoch = args.max_epoch
    early_stop = args.early_stop
    multi_gpu = args.multi_gpu
    wandb = args.wandb
    num_workers = args.num_workers

    if architecture is None:
        raise ValueError("A model architecture has to be specified: " + str(architectures))

    if transformer is None and embeddings_name is None:
        # default word embeddings
        embeddings_name = "glove-840B"

    if action == Tasks.TRAIN:
        train(
            model,
            embeddings_name=embeddings_name,
            architecture=architecture,
            transformer=transformer,
            input_path=input_path,
            output_path=output,
            max_sequence_length=max_sequence_length,
            batch_size=batch_size,
            incremental=incremental,
            input_model_path=input_model_path,
            patience=patience,
            learning_rate=learning_rate,
            max_epoch=max_epoch,
            early_stop=early_stop,
            multi_gpu=multi_gpu,
            report_to_wandb=wandb,
            num_workers=num_workers,
        )

    if action == Tasks.EVAL:
        if args.fold_count is not None and args.fold_count > 1:
            print(
                "The argument fold-count argument will be ignored. For n-fold cross-validation, please use "
                "it in combination with " + str(Tasks.TRAIN_EVAL)
            )
        if input_path is None:
            raise ValueError(
                "A Grobid evaluation data file must be specified to evaluate a grobid model with the parameter --input"
            )
        eval_(model, input_path=input_path, architecture=architecture)

    if action == Tasks.TRAIN_EVAL:
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")
        train_eval(
            model,
            embeddings_name=embeddings_name,
            architecture=architecture,
            transformer=transformer,
            input_path=input_path,
            output_path=output,
            fold_count=args.fold_count,
            max_sequence_length=max_sequence_length,
            batch_size=batch_size,
            incremental=incremental,
            input_model_path=input_model_path,
            learning_rate=learning_rate,
            max_epoch=max_epoch,
            early_stop=early_stop,
            multi_gpu=multi_gpu,
            report_to_wandb=wandb,
        )

    if action == Tasks.TAG:
        someTexts = []

        if model == "date":
            someTexts.append("January 2006")
            someTexts.append("March the 27th, 2001")
            someTexts.append(" on  April 27, 2001. . ")
            someTexts.append("2018")
            someTexts.append("2023 July the 22nd")
        elif model == "citation":
            someTexts.append(
                'N. Al-Dhahir and J. Cioffi, "On the uniform ADC bit precision and clip level computation for a Gaussian signal," IEEE Trans. Signal Processing, pp. 434–438, Feb. 1996.'
            )
            someTexts.append(
                "T. Steinherz, E. Rivlin, N. Intrator, Off-line cursive script word recognition—a survey, Int. J. Doc. Anal. Recognition 2(3) (1999) 1-33."
            )
        elif model == "name-citation":
            someTexts.append("L. Romary and L. Foppiano")
            someTexts.append("Maniscalco, S., Francica, F., Zaffino, R.L.")
        elif model == "name-header":
            someTexts.append("He-Jin Wu 1 · Zhao Jin 2 · Ai-Dong Zhu 1")
            someTexts.append("Irène Charon ⋆ and Olivier Hudry")
        elif model == "software":
            someTexts.append(
                "The column scores (the fraction of entirely correct columns) were  reported  in  addition  to Q-scores  for  BAliBASE 3.0. Wilcoxon  signed-ranks  tests  were  performed  to  calculate statistical  significance  of  comparisons  between  alignment programs,   which   include   ProbCons   (version   1.10)   (23), MAFFT (version 5.667) (11) with several options, MUSCLE (version 3.52) (10) and ClustalW (version 1.83) (7)."
            )
            someTexts.append(
                "Wilcoxon signed-ranks tests were performed to calculate statistical significance of comparisons between  alignment programs, which include ProbCons (version 1.10) (23) MAFFT (version 5.667) (11) with several options, MUSCLE (version 3.52) (10) and ClustalW (version 1.83) (7)."
            )
            someTexts.append(
                "All statistical analyses were done using computer software Prism 6 for Windows (version 6.02; GraphPad Software, San Diego, CA, USA). One-Way ANOVA was used to detect differences amongst the groups. To account for the non-normal distribution of the data, all data were sorted by rank status prior to ANOVA statistical analysis. "
            )
            someTexts.append(
                "The statistical analysis was performed using IBM SPSS Statistics v. 20 (SPSS Inc, 2003, Chicago, USA)."
            )

        if architecture.find("FEATURE") == -1:
            result = annotate_text(someTexts, model, "json", architecture=architecture, multi_gpu=multi_gpu)
            print(json.dumps(result, sort_keys=False, indent=4, ensure_ascii=False))
        else:
            print(
                "The model " + architecture + " cannot be used without supplying features as input and it's disabled. "
                "Please supply an architecture without features. "
            )



from typing import List

import numpy as np

__author__ = "@de-code"

"""
Utility class, from:
https://github.com/elifesciences/sciencebeam-trainer-delft/blob/develop/sciencebeam_trainer_delft/utils/numpy.py
"""


def concatenate_or_none(arrays: List[np.array], **kwargs) -> np.array:
    if arrays[0] is None:
        return None
    return np.concatenate(arrays, **kwargs)


# https://stackoverflow.com/a/51526109/8676953
def shuffle_arrays(arrays: List[np.array], random_seed: int = None) -> List[np.array]:
    """Shuffles arrays in-place, in the same order, along axis=0

    Parameters:
    -----------
    arrays : List of NumPy arrays.
    random_seed : Seed value if not None, else seed is random.
    """
    assert all(len(arr) == len(arrays[0]) for arr in arrays)
    if random_seed is None:
        random_seed = np.random.randint(0, 2 ** (32 - 1) - 1)

    for arr in arrays:
        rstate = np.random.RandomState(random_seed)  # pylint: disable=no-member
        rstate.shuffle(arr)


def shuffle_triple_with_view(a, b=None, c=None):
    """
    Shuffle a pair/triple with view, without modifying the initial elements.
    The assumption is that if b is None, c is also None.
    """
    if b is not None and c is not None:
        assert (
            "Cannot shuffle with view if the arrays have different dimensions: " + str(len(a)) + " vs " + str(len(b))
        ), len(a) == len(b) == len(c)
    elif b is not None and c is None:
        assert (
            "Cannot shuffle with view if the arrays have different dimensions: " + str(len(a)) + " vs " + str(len(b))
        ), len(a) == len(b)
    elif c is not None and b is None:
        raise Exception("b is None but c is not.")

    # generate permutation index array
    permutation = np.random.permutation(a.shape[0])

    # shuffle the arrays
    if c is not None and b is not None:
        return a[permutation], b[permutation], c[permutation]
    if b is not None and c is None:
        return a[permutation], b[permutation], None
    else:
        return a[permutation], None, None



from delft.utilities.misc import parse_dict, parse_number_ranges

# derived from https://github.com/elifesciences/sciencebeam-trainer-delft/tree/develop/tests


class TestParseNumberRanges:
    def test_should_return_empty_array_for_empty_expr(self):
        assert parse_number_ranges("") == []

    def test_should_parse_single_number(self):
        assert parse_number_ranges("1") == [1]

    def test_should_parse_comma_separated_numbers(self):
        assert parse_number_ranges("1,2,3") == [1, 2, 3]

    def test_should_parse_single_number_range(self):
        assert parse_number_ranges("1-3") == [1, 2, 3]

    def test_should_parse_multiple_number_ranges(self):
        assert parse_number_ranges("1-3,5-6") == [1, 2, 3, 5, 6]

    def test_should_ignore_spaces(self):
        assert parse_number_ranges(" 1 - 3 , 5 - 6 ") == [1, 2, 3, 5, 6]


class TestParseDict:
    def test_should_return_empty_dict_for_empty_expr(self):
        assert parse_dict("") == {}

    def test_should_parse_single_key_value_pair(self):
        assert parse_dict("key1=value1") == {"key1": "value1"}

    def test_should_allow_equals_sign_in_value(self):
        assert parse_dict("key1=value=1") == {"key1": "value=1"}

    def test_should_parse_multiple_key_value_pair(self):
        assert parse_dict("key1=value1|key2=value2", delimiter="|") == {"key1": "value1", "key2": "value2"}

    def test_should_ignore_spaces(self):
        assert parse_dict(" key1 = value1 | key2 = value2 ", delimiter="|") == {"key1": "value1", "key2": "value2"}



import argparse
import json
import time

from delft.textClassification import Classifier
from delft.textClassification.models import architectures
from delft.textClassification.reader import load_citation_sentiment_corpus
from delft.utilities.Utilities import split_data_and_labels

list_classes = ["negative", "neutral", "positive"]

class_weights = {0: 25.0, 1: 1.0, 2: 9.0}


def configure(architecture):
    batch_size = 256
    maxlen = 150
    patience = 5
    early_stop = True
    max_epoch = 60

    # default bert model parameters
    if architecture == "bert":
        batch_size = 32
        early_stop = False
        max_epoch = 3

    return batch_size, maxlen, patience, early_stop, max_epoch


def train(embeddings_name, fold_count, architecture="gru", transformer=None):
    batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)

    model = Classifier(
        "citations_" + architecture,
        architecture=architecture,
        list_classes=list_classes,
        max_epoch=max_epoch,
        fold_number=fold_count,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        patience=patience,
        early_stop=early_stop,
        class_weights=class_weights,
        transformer_name=transformer,
    )

    print("loading citation sentiment corpus...")
    xtr, y = load_citation_sentiment_corpus("data/textClassification/citations/citation_sentiment_corpus.txt")

    if fold_count == 1:
        model.train(xtr, y)
    else:
        model.train_nfold(xtr, y)
    # saving the model
    model.save()


def train_and_eval(embeddings_name, fold_count, architecture="gru", transformer=None):
    batch_size, maxlen, patience, early_stop, max_epoch = configure(architecture)

    model = Classifier(
        "citations_" + architecture,
        architecture=architecture,
        list_classes=list_classes,
        max_epoch=max_epoch,
        fold_number=fold_count,
        use_roc_auc=True,
        embeddings_name=embeddings_name,
        batch_size=batch_size,
        maxlen=maxlen,
        patience=patience,
        early_stop=early_stop,
        class_weights=class_weights,
        transformer_name=transformer,
    )

    print("loading citation sentiment corpus...")
    xtr, y = load_citation_sentiment_corpus("data/textClassification/citations/citation_sentiment_corpus.txt")

    # segment train and eval sets
    x_train, y_train, x_test, y_test = split_data_and_labels(xtr, y, 0.9)

    if fold_count == 1:
        model.train(x_train, y_train)
    else:
        model.train_nfold(x_train, y_train)

    # saving the model
    model.save()

    model.eval(x_test, y_test)


# classify a list of texts
def classify(texts, output_format, architecture="gru", embeddings_name=None, transformer=None):
    # load model
    model = Classifier(
        "citations_" + architecture,
        architecture=architecture,
        list_classes=list_classes,
        embeddings_name=embeddings_name,
        transformer_name=transformer,
    )
    model.load()
    start_time = time.time()
    result = model.predict(texts, output_format)
    runtime = round(time.time() - start_time, 3)
    if output_format == "json":
        result["runtime"] = runtime
    else:
        print("runtime: %s seconds " % (runtime))
    return result


if __name__ == "__main__":
    parser = argparse.ArgumentParser(description="Sentiment classification of citation contexts based on DeLFT")

    word_embeddings_examples = ["glove-840B", "fasttext-crawl", "word2vec"]
    pretrained_transformers_examples = ["bert-base-cased", "bert-large-cased", "allenai/scibert_scivocab_cased"]

    parser.add_argument("action", help="one of [train, train_eval, classify]")
    parser.add_argument("--fold-count", type=int, default=1)
    parser.add_argument(
        "--architecture", default="gru", help="type of model architecture to be used, one of " + str(architectures)
    )
    parser.add_argument(
        "--embedding",
        default=None,
        help="The desired pre-trained word embeddings using their descriptions in the file. "
        + "For local loading, use delft/resources-registry.json. "
        + "Be sure to use here the same name as in the registry, e.g. "
        + str(word_embeddings_examples)
        + " and that the path in the registry to the embedding file is correct on your system.",
    )
    parser.add_argument(
        "--transformer",
        default=None,
        help="The desired pre-trained transformer to be used in the selected architecture. "
        + "For local loading use, delft/resources-registry.json, and be sure to use here the same name as in the registry, e.g. "
        + str(pretrained_transformers_examples)
        + " and that the path in the registry to the model path is correct on your system. "
        + "HuggingFace transformers hub will be used otherwise to fetch the model, see https://huggingface.co/models "
        + "for model names",
    )

    args = parser.parse_args()

    if args.action not in ("train", "train_eval", "classify"):
        print("action not specifed, must be one of [train,train_eval,classify]")

    embeddings_name = args.embedding
    transformer = args.transformer

    architecture = args.architecture
    if architecture not in architectures:
        print("unknown model architecture, must be one of " + str(architectures))

    if transformer is None and embeddings_name is None:
        # default word embeddings
        embeddings_name = "glove-840B"

    if args.action == "train":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")

        train(embeddings_name, args.fold_count, architecture=architecture, transformer=transformer)

    if args.action == "train_eval":
        if args.fold_count < 1:
            raise ValueError("fold-count should be equal or more than 1")

        y_test = train_and_eval(embeddings_name, args.fold_count, architecture=architecture, transformer=transformer)

    if args.action == "classify":
        someTexts = [
            "One successful strategy [15] computes the set-similarity involving (multi-word) keyphrases about the mentions and the entities, collected from the KG.",
            "Unfortunately, fewer than half of the OCs in the DAML02 OC catalog (Dias et al. 2002) are suitable for use with the isochrone-fitting method because of the lack of a prominent main sequence, in addition to an absence of radial velocity and proper-motion data.",
            "However, we found that the pairwise approach LambdaMART [41] achieved the best performance on our datasets among most learning to rank algorithms.",
        ]
        result = classify(
            someTexts, "json", architecture=architecture, embeddings_name=embeddings_name, transformer=transformer
        )
        print(json.dumps(result, sort_keys=False, indent=4, ensure_ascii=False))



from tf_keras.layers import (
    GRU,
    LSTM,
    Bidirectional,
    Concatenate,
    Conv1D,
    Dense,
    Dropout,
    Embedding,
    GlobalMaxPooling1D,
    Input,
    TimeDistributed,
)
from tf_keras.models import Model, clone_model

from delft.sequenceLabelling.config import ModelConfig
from delft.sequenceLabelling.data_generator import DataGenerator, DataGeneratorTransformers
from delft.sequenceLabelling.preprocess import BERTPreprocessor, Preprocessor
from delft.utilities.crf_layer import ChainCRF
from delft.utilities.crf_wrapper_default import CRFModelWrapperDefault
from delft.utilities.crf_wrapper_for_bert import CRFModelWrapperForBERT
from delft.utilities.Embeddings import load_resource_registry
from delft.utilities.Transformer import Transformer

"""
The sequence labeling models.

Each architecture model is a class implementing a Keras architecture.
The architecture class can also define the data generator class object to be used, the loss function,
the metrics and the optimizer.
"""


def get_model(config: ModelConfig, preprocessor, ntags=None, load_pretrained_weights=True, local_path=None):
    """
    Return a model instance by its name. This is a facilitator function.
    """
    print(config.architecture)

    if config.architecture == BidLSTM.name:
        preprocessor.return_word_embeddings = True
        preprocessor.return_chars = True
        preprocessor.return_lengths = True
        return BidLSTM(config, ntags)

    elif config.architecture == BidLSTM_CRF.name:
        preprocessor.return_word_embeddings = True
        preprocessor.return_chars = True
        preprocessor.return_lengths = True
        config.use_crf = True
        return BidLSTM_CRF(config, ntags)

    elif config.architecture == BidLSTM_ChainCRF.name:
        preprocessor.return_word_embeddings = True
        preprocessor.return_chars = True
        preprocessor.return_lengths = True
        config.use_crf = True
        config.use_chain_crf = True
        return BidLSTM_ChainCRF(config, ntags)

    elif config.architecture == BidLSTM_CNN.name:
        preprocessor.return_word_embeddings = True
        preprocessor.return_casing = True
        preprocessor.return_chars = True
        preprocessor.return_lengths = True
        return BidLSTM_CNN(config, ntags)

    elif config.architecture == BidLSTM_CNN_CRF.name:
        preprocessor.return_word_embeddings = True
        preprocessor.return_casing = True
        preprocessor.return_chars = True
        preprocessor.return_lengths = True
        config.use_crf = True
        return BidLSTM_CNN_CRF(config, ntags)

    elif config.architecture == BidGRU_CRF.name:
        preprocessor.return_word_embeddings = True
        preprocessor.return_chars = True
        preprocessor.return_lengths = True
        config.use_crf = True
        return BidGRU_CRF(config, ntags)

    elif config.architecture == BidLSTM_CRF_FEATURES.name:
        preprocessor.return_word_embeddings = True
        preprocessor.return_features = True
        preprocessor.return_chars = True
        preprocessor.return_lengths = True
        config.use_crf = True
        return BidLSTM_CRF_FEATURES(config, ntags)

    elif config.architecture == BidLSTM_ChainCRF_FEATURES.name:
        preprocessor.return_word_embeddings = True
        preprocessor.return_features = True
        preprocessor.return_chars = True
        preprocessor.return_lengths = True
        config.use_crf = True
        config.use_chain_crf = True
        return BidLSTM_ChainCRF_FEATURES(config, ntags)

    elif config.architecture == BidLSTM_CRF_CASING.name:
        preprocessor.return_word_embeddings = True
        preprocessor.return_casing = True
        preprocessor.return_chars = True
        preprocessor.return_lengths = True
        config.use_crf = True
        return BidLSTM_CRF_CASING(config, ntags)

    elif config.architecture == BERT.name:
        preprocessor.return_bert_embeddings = True
        config.labels = preprocessor.vocab_tag
        return BERT(
            config,
            ntags,
            load_pretrained_weights=load_pretrained_weights,
            local_path=local_path,
            preprocessor=preprocessor,
        )

    elif config.architecture == BERT_FEATURES.name:
        preprocessor.return_bert_embeddings = True
        preprocessor.return_features = True
        config.labels = preprocessor.vocab_tag
        return BERT_FEATURES(
            config,
            ntags,
            load_pretrained_weights=load_pretrained_weights,
            local_path=local_path,
            preprocessor=preprocessor,
        )

    elif config.architecture == BERT_CRF.name:
        preprocessor.return_bert_embeddings = True
        config.use_crf = True
        config.labels = preprocessor.vocab_tag
        return BERT_CRF(
            config,
            ntags,
            load_pretrained_weights=load_pretrained_weights,
            local_path=local_path,
            preprocessor=preprocessor,
        )

    elif config.architecture == BERT_ChainCRF.name:
        preprocessor.return_bert_embeddings = True
        config.use_crf = True
        config.use_chain_crf = True
        config.labels = preprocessor.vocab_tag
        return BERT_ChainCRF(
            config,
            ntags,
            load_pretrained_weights=load_pretrained_weights,
            local_path=local_path,
            preprocessor=preprocessor,
        )

    elif config.architecture == BERT_CRF_FEATURES.name:
        preprocessor.return_bert_embeddings = True
        preprocessor.return_features = True
        config.use_crf = True
        config.labels = preprocessor.vocab_tag
        return BERT_CRF_FEATURES(
            config,
            ntags,
            load_pretrained_weights=load_pretrained_weights,
            local_path=local_path,
            preprocessor=preprocessor,
        )

    elif config.architecture == BERT_ChainCRF_FEATURES.name:
        preprocessor.return_bert_embeddings = True
        preprocessor.return_features = True
        config.use_crf = True
        config.use_chain_crf = True
        config.labels = preprocessor.vocab_tag
        return BERT_ChainCRF_FEATURES(
            config,
            ntags,
            load_pretrained_weights=load_pretrained_weights,
            local_path=local_path,
            preprocessor=preprocessor,
        )

    elif config.architecture == BERT_CRF_CHAR.name:
        preprocessor.return_bert_embeddings = True
        preprocessor.return_chars = True
        config.use_crf = True
        config.labels = preprocessor.vocab_tag
        return BERT_CRF_CHAR(
            config,
            ntags,
            load_pretrained_weights=load_pretrained_weights,
            local_path=local_path,
            preprocessor=preprocessor,
        )

    elif config.architecture == BERT_CRF_CHAR_FEATURES.name:
        preprocessor.return_bert_embeddings = True
        preprocessor.return_features = True
        preprocessor.return_chars = True
        config.use_crf = True
        config.labels = preprocessor.vocab_tag
        return BERT_CRF_CHAR_FEATURES(
            config,
            ntags,
            load_pretrained_weights=load_pretrained_weights,
            local_path=local_path,
            preprocessor=preprocessor,
        )
    else:
        raise (OSError("Model name does exist: " + config.architecture))


class BaseModel(object):
    """
    Base class for DeLFT sequence labeling models

    Args:
        config (ModelConfig): DeLFT model configuration object
        ntags (integer): number of different labels of the model
        load_pretrained_weights (boolean): used only when the model contains a transformer layer - indicate whether
                                           or not we load the pretrained weights of this transformer. For training
                                           a new model set it to True. When getting the full Keras model to load
                                           existing weights, set it False to avoid reloading the pretrained weights.
        local_path (string): used only when the model contains a transformer layer - the path where to load locally the
                             pretrained transformer. If None, the transformer model will be fetched from HuggingFace
                             transformers hub.
    """

    transformer_config = None
    transformer_preprocessor = None

    def __init__(
        self, config, ntags=None, load_pretrained_weights: bool = True, local_path: str = None, preprocessor=None
    ):
        self.config = config
        self.ntags = ntags
        self.model = None
        self.local_path = local_path
        self.load_pretrained_weights = load_pretrained_weights

        self.registry = load_resource_registry("delft/resources-registry.json")

    def predict(self, X, *args, **kwargs):
        y_pred = self.model.predict(X, batch_size=1)
        return y_pred

    def evaluate(self, X, y):
        score = self.model.evaluate(X, y, batch_size=1)
        return score

    def save(self, filepath):
        self.model.save_weights(filepath)

    def load(self, filepath):
        print("loading model weights", filepath)
        self.model.load_weights(filepath=filepath)

    def __getattr__(self, name):
        return getattr(self.model, name)

    def clone_model(self):
        model_copy = clone_model(self.model)
        model_copy.set_weights(self.model.get_weights())
        return model_copy

    def get_generator(self):
        # default generator
        return DataGenerator

    def print_summary(self):
        if hasattr(self.model, "base_model"):
            self.model.base_model.summary()
        self.model.summary()

    def init_transformer(
        self, config: ModelConfig, load_pretrained_weights: bool, local_path: str, preprocessor: Preprocessor
    ):
        transformer = Transformer(config.transformer_name, resource_registry=self.registry, delft_local_path=local_path)
        print(config.transformer_name, "will be used, loaded via", transformer.loading_method)
        transformer_model = transformer.instantiate_layer(load_pretrained_weights=load_pretrained_weights)
        self.transformer_config = transformer.transformer_config
        transformer.init_preprocessor(max_sequence_length=config.max_sequence_length)

        self.transformer_preprocessor = BERTPreprocessor(
            transformer.tokenizer, preprocessor.empty_features_vector(), preprocessor.empty_char_vector()
        )

        return transformer_model


class BidLSTM(BaseModel):
    """
    A Keras implementation of simple BidLSTM for sequence labelling with character and word inputs, and softmax final layer.
    """

    name = "BidLSTM"

    def __init__(self, config, ntags=None):
        super().__init__(config, ntags)

        # build input, directly feed with word embedding by the data generator
        word_input = Input(shape=(None, config.word_embedding_size), name="word_input")

        # build character based embedding
        char_input = Input(shape=(None, config.max_char_length), dtype="int32", name="char_input")
        char_embeddings = TimeDistributed(
            Embedding(
                input_dim=config.char_vocab_size,
                output_dim=config.char_embedding_size,
                mask_zero=True,
                # embeddings_initializer=RandomUniform(minval=-0.5, maxval=0.5),
                name="char_embeddings",
            )
        )(char_input)

        chars = TimeDistributed(Bidirectional(LSTM(config.num_char_lstm_units, return_sequences=False)))(
            char_embeddings
        )

        # length of sequence not used by the model, but used by the training scorer
        length_input = Input(batch_shape=(None, 1), dtype="int32", name="length_input")

        # combine characters and word embeddings
        x = Concatenate()([word_input, chars])
        x = Dropout(config.dropout)(x)

        x = Bidirectional(
            LSTM(units=config.num_word_lstm_units, return_sequences=True, recurrent_dropout=config.recurrent_dropout)
        )(x)
        x = Dropout(config.dropout)(x)
        pred = Dense(ntags, activation="softmax")(x)

        self.model = Model(inputs=[word_input, char_input, length_input], outputs=[pred])
        # self.model.summary()
        self.config = config


class BidLSTM_CRF(BaseModel):
    """
    A Keras implementation of BidLSTM-CRF for sequence labelling.

    References
    --
    Guillaume Lample, Miguel Ballesteros, Sandeep Subramanian, Kazuya Kawakami, Chris Dyer.
    "Neural Architectures for Named Entity Recognition". Proceedings of NAACL 2016.
    https://arxiv.org/abs/1603.01360
    """

    name = "BidLSTM_CRF"

    def __init__(self, config, ntags=None):
        super().__init__(config, ntags)

        # build input, directly feed with word embedding by the data generator
        word_input = Input(shape=(None, config.word_embedding_size), name="word_input")

        # build character based embedding
        char_input = Input(shape=(None, config.max_char_length), dtype="int32", name="char_input")
        char_embeddings = TimeDistributed(
            Embedding(
                input_dim=config.char_vocab_size,
                output_dim=config.char_embedding_size,
                mask_zero=True,
                # embeddings_initializer=RandomUniform(minval=-0.5, maxval=0.5),
                name="char_embeddings",
            )
        )(char_input)

        chars = TimeDistributed(Bidirectional(LSTM(config.num_char_lstm_units, return_sequences=False)))(
            char_embeddings
        )

        # length of sequence not used by the model, but used by the training scorer
        length_input = Input(batch_shape=(None, 1), dtype="int32", name="length_input")

        # combine characters and word embeddings
        x = Concatenate()([word_input, chars])
        x = Dropout(config.dropout)(x)

        x = Bidirectional(
            LSTM(units=config.num_word_lstm_units, return_sequences=True, recurrent_dropout=config.recurrent_dropout)
        )(x)
        x = Dropout(config.dropout)(x)
        x = Dense(config.num_word_lstm_units, activation="tanh")(x)

        base_model = Model(inputs=[word_input, char_input, length_input], outputs=[x])

        self.model = CRFModelWrapperDefault(base_model, ntags)
        self.model.build(
            input_shape=[
                (None, None, config.word_embedding_size),
                (None, None, config.max_char_length),
                (None, None, 1),
            ]
        )
        # self.model.summary()
        self.config = config


class BidLSTM_ChainCRF(BaseModel):
    """
    A Keras implementation of BidLSTM-CRF for sequence labelling with an alternative CRF layer implementation.

    References
    --
    Guillaume Lample, Miguel Ballesteros, Sandeep Subramanian, Kazuya Kawakami, Chris Dyer.
    "Neural Architectures for Named Entity Recognition". Proceedings of NAACL 2016.
    https://arxiv.org/abs/1603.01360
    """

    name = "BidLSTM_ChainCRF"

    def __init__(self, config, ntags=None):
        super().__init__(config, ntags)

        # build input, directly feed with word embedding by the data generator
        word_input = Input(shape=(None, config.word_embedding_size), name="word_input")

        # build character based embedding
        char_input = Input(shape=(None, config.max_char_length), dtype="int32", name="char_input")
        char_embeddings = TimeDistributed(
            Embedding(
                input_dim=config.char_vocab_size,
                output_dim=config.char_embedding_size,
                mask_zero=False,
                # embeddings_initializer=RandomUniform(minval=-0.5, maxval=0.5),
                name="char_embeddings",
            )
        )(char_input)

        chars = TimeDistributed(Bidirectional(LSTM(config.num_char_lstm_units, return_sequences=False)))(
            char_embeddings
        )

        # length of sequence not used by the model, but used by the training scorer
        length_input = Input(batch_shape=(None, 1), dtype="int32", name="length_input")

        # combine characters and word embeddings
        x = Concatenate()([word_input, chars])
        x = Dropout(config.dropout)(x)

        x = Bidirectional(
            LSTM(units=config.num_word_lstm_units, return_sequences=True, recurrent_dropout=config.recurrent_dropout)
        )(x)
        x = Dropout(config.dropout)(x)
        x = Dense(config.num_word_lstm_units, activation="tanh")(x)
        x = Dense(ntags)(x)
        self.crf = ChainCRF()
        pred = self.crf(x)

        self.model = Model(inputs=[word_input, char_input, length_input], outputs=[pred])
        self.config = config


class BidLSTM_CNN(BaseModel):
    """
    A Keras implementation of BidLSTM-CNN for sequence labelling.

    References
    --
    Jason P. C. Chiu, Eric Nichols. "Named Entity Recognition with Bidirectional LSTM-CNNs". 2016.
    https://arxiv.org/abs/1511.08308
    """

    name = "BidLSTM_CNN"

    def __init__(self, config, ntags=None):
        super().__init__(config, ntags)

        # build input, directly feed with word embedding by the data generator
        word_input = Input(shape=(None, config.word_embedding_size), name="word_input")

        # build character based embedding
        char_input = Input(shape=(None, config.max_char_length), dtype="int32", name="char_input")
        char_embeddings = TimeDistributed(
            Embedding(
                input_dim=config.char_vocab_size,
                output_dim=config.char_embedding_size,
                mask_zero=False,
                name="char_embeddings",
            )
        )(char_input)

        dropout = Dropout(config.dropout)(char_embeddings)

        conv1d_out = TimeDistributed(Conv1D(kernel_size=3, filters=30, padding="same", activation="tanh", strides=1))(
            dropout
        )
        maxpool_out = TimeDistributed(GlobalMaxPooling1D())(conv1d_out)
        chars = Dropout(config.dropout)(maxpool_out)

        # custom features input and embeddings
        casing_input = Input(
            batch_shape=(
                None,
                None,
            ),
            dtype="int32",
            name="casing_input",
        )
        casing_embedding = Embedding(
            input_dim=config.case_vocab_size,
            output_dim=config.case_embedding_size,
            # mask_zero=True,
            trainable=False,
            name="casing_embedding",
        )(casing_input)
        casing_embedding = Dropout(config.dropout)(casing_embedding)

        # length of sequence not used by the model, but used by the training scorer
        length_input = Input(batch_shape=(None, 1), dtype="int32")

        # combine words, custom features and characters
        x = Concatenate(axis=-1)([word_input, casing_embedding, chars])
        x = Dropout(config.dropout)(x)
        x = Bidirectional(
            LSTM(units=config.num_word_lstm_units, return_sequences=True, recurrent_dropout=config.recurrent_dropout)
        )(x)
        x = Dropout(config.dropout)(x)
        # pred = TimeDistributed(Dense(ntags, activation='softmax'))(x)
        pred = Dense(ntags, activation="softmax")(x)

        self.model = Model(inputs=[word_input, char_input, casing_input, length_input], outputs=[pred])
        self.config = config


class BidLSTM_CNN_CRF(BaseModel):
    """
    A Keras implementation of BidLSTM-CNN-CRF for sequence labelling.

    References
    --
    Xuezhe Ma and Eduard Hovy. "End-to-end Sequence Labeling via Bi-directional LSTM-CNNs-CRF". 2016.
    https://arxiv.org/abs/1603.01354
    """

    name = "BidLSTM_CNN_CRF"

    def __init__(self, config, ntags=None):
        super().__init__(config, ntags)

        # build input, directly feed with word embedding by the data generator
        word_input = Input(shape=(None, config.word_embedding_size), name="word_input")

        # build character based embedding
        char_input = Input(shape=(None, config.max_char_length), dtype="int32", name="char_input")
        char_embeddings = TimeDistributed(
            Embedding(
                input_dim=config.char_vocab_size,
                output_dim=config.char_embedding_size,
                mask_zero=False,
                name="char_embeddings",
            )
        )(char_input)

        dropout = Dropout(config.dropout)(char_embeddings)

        conv1d_out = TimeDistributed(Conv1D(kernel_size=3, filters=30, padding="same", activation="tanh", strides=1))(
            dropout
        )
        maxpool_out = TimeDistributed(GlobalMaxPooling1D())(conv1d_out)
        chars = Dropout(config.dropout)(maxpool_out)

        # custom features input and embeddings
        casing_input = Input(
            batch_shape=(
                None,
                None,
            ),
            dtype="int32",
            name="casing_input",
        )

        """
        casing_embedding = Embedding(input_dim=config.case_vocab_size,
                           output_dim=config.case_embedding_size,
                           mask_zero=True,
                           trainable=False,
                           name='casing_embedding')(casing_input)
        casing_embedding = Dropout(config.dropout)(casing_embedding)
        """

        # length of sequence not used by the model, but used by the training scorer
        length_input = Input(batch_shape=(None, 1), dtype="int32")

        # combine words, custom features and characters
        x = Concatenate(axis=-1)([word_input, chars])
        x = Dropout(config.dropout)(x)

        x = Bidirectional(
            LSTM(units=config.num_word_lstm_units, return_sequences=True, recurrent_dropout=config.recurrent_dropout)
        )(x)
        x = Dropout(config.dropout)(x)
        x = Dense(config.num_word_lstm_units, activation="tanh")(x)

        base_model = Model(inputs=[word_input, char_input, casing_input, length_input], outputs=[x])
        self.model = CRFModelWrapperDefault(base_model, ntags)
        self.model.build(
            input_shape=[
                (None, None, config.word_embedding_size),
                (None, None, config.max_char_length),
                (None, None, None),
                (None, None, 1),
            ]
        )
        self.config = config


class BidGRU_CRF(BaseModel):
    """
    A Keras implementation of BidGRU-CRF for sequence labelling.
    """

    name = "BidGRU_CRF"

    def __init__(self, config, ntags=None):
        super().__init__(config, ntags)

        # build input, directly feed with word embedding by the data generator
        word_input = Input(shape=(None, config.word_embedding_size), name="word_input")

        # build character based embedding
        char_input = Input(shape=(None, config.max_char_length), dtype="int32", name="char_input")
        char_embeddings = TimeDistributed(
            Embedding(
                input_dim=config.char_vocab_size,
                output_dim=config.char_embedding_size,
                mask_zero=True,
                # embeddings_initializer=RandomUniform(minval=-0.5, maxval=0.5),
                name="char_embeddings",
            )
        )(char_input)

        chars = TimeDistributed(Bidirectional(LSTM(config.num_char_lstm_units, return_sequences=False)))(
            char_embeddings
        )

        # length of sequence not used by the model, but used by the training scorer
        length_input = Input(batch_shape=(None, 1), dtype="int32", name="length_input")

        # combine characters and word embeddings
        x = Concatenate()([word_input, chars])
        x = Dropout(config.dropout)(x)

        x = Bidirectional(
            GRU(units=config.num_word_lstm_units, return_sequences=True, recurrent_dropout=config.recurrent_dropout)
        )(x)
        x = Dropout(config.dropout)(x)
        x = Bidirectional(
            GRU(units=config.num_word_lstm_units, return_sequences=True, recurrent_dropout=config.recurrent_dropout)
        )(x)
        x = Dense(config.num_word_lstm_units, activation="tanh")(x)

        base_model = Model(inputs=[word_input, char_input, length_input], outputs=[x])
        self.model = CRFModelWrapperDefault(base_model, ntags)
        self.model.build(
            input_shape=[
                (None, None, config.word_embedding_size),
                (None, None, config.max_char_length),
                (None, None, 1),
            ]
        )

        self.config = config


class BidLSTM_CRF_CASING(BaseModel):
    """
    A Keras implementation of BidLSTM-CRF for sequence labelling with additinal features related to casing
    (inferred from word forms).
    """

    name = "BidLSTM_CRF_CASING"

    def __init__(self, config, ntags=None):
        super().__init__(config, ntags)

        # build input, directly feed with word embedding by the data generator
        word_input = Input(shape=(None, config.word_embedding_size), name="word_input")

        # build character based embedding
        char_input = Input(shape=(None, config.max_char_length), dtype="int32", name="char_input")
        char_embeddings = TimeDistributed(
            Embedding(
                input_dim=config.char_vocab_size,
                output_dim=config.char_embedding_size,
                mask_zero=True,
                # embeddings_initializer=RandomUniform(minval=-0.5, maxval=0.5),
                name="char_embeddings",
            )
        )(char_input)

        chars = TimeDistributed(Bidirectional(LSTM(config.num_char_lstm_units, return_sequences=False)))(
            char_embeddings
        )

        # custom features input and embeddings
        casing_input = Input(
            batch_shape=(
                None,
                None,
            ),
            dtype="int32",
            name="casing_input",
        )

        casing_embedding = Embedding(
            input_dim=config.case_vocab_size,
            output_dim=config.case_embedding_size,
            # mask_zero=True,
            trainable=False,
            name="casing_embedding",
        )(casing_input)
        casing_embedding = Dropout(config.dropout)(casing_embedding)

        # length of sequence not used by the model, but used by the training scorer
        length_input = Input(batch_shape=(None, 1), dtype="int32", name="length_input")

        # combine characters and word embeddings
        x = Concatenate()([word_input, casing_embedding, chars])
        x = Dropout(config.dropout)(x)

        x = Bidirectional(
            LSTM(units=config.num_word_lstm_units, return_sequences=True, recurrent_dropout=config.recurrent_dropout)
        )(x)
        x = Dropout(config.dropout)(x)
        x = Dense(config.num_word_lstm_units, activation="tanh")(x)

        base_model = Model(inputs=[word_input, char_input, casing_input, length_input], outputs=[x])
        self.model = CRFModelWrapperDefault(base_model, ntags)
        self.model.build(
            input_shape=[
                (None, None, config.word_embedding_size),
                (None, None, config.max_char_length),
                (None, None),
                (None, None, 1),
            ]
        )
        self.config = config


class BidLSTM_CRF_FEATURES(BaseModel):
    """
    A Keras implementation of BidLSTM-CRF for sequence labelling using tokens combined with
    additional generic discrete features information.
    """

    name = "BidLSTM_CRF_FEATURES"

    def __init__(self, config, ntags=None):
        super().__init__(config, ntags)

        # build input, directly feed with word embedding by the data generator
        word_input = Input(shape=(None, config.word_embedding_size), name="word_input")

        # build character based embedding
        char_input = Input(shape=(None, config.max_char_length), dtype="int32", name="char_input")
        char_embeddings = TimeDistributed(
            Embedding(
                input_dim=config.char_vocab_size,
                output_dim=config.char_embedding_size,
                mask_zero=True,
                # embeddings_initializer=RandomUniform(minval=-0.5, maxval=0.5),
                name="char_embeddings",
            )
        )(char_input)

        chars = TimeDistributed(Bidirectional(LSTM(config.num_char_lstm_units, return_sequences=False)))(
            char_embeddings
        )

        # layout features input and embeddings
        features_input = Input(shape=(None, len(config.features_indices)), dtype="float32", name="features_input")

        # The input dimension is calculated by
        # features_vocabulary_size (default 12) * number_of_features + 1 (the zero is reserved for masking / padding)
        features_embedding = TimeDistributed(
            Embedding(
                input_dim=config.features_vocabulary_size * len(config.features_indices) + 1,
                output_dim=config.features_embedding_size,
                # mask_zero=True,
                trainable=True,
                name="features_embedding",
            ),
            name="features_embedding_td",
        )(features_input)

        features_embedding_bd = TimeDistributed(
            Bidirectional(LSTM(config.features_lstm_units, return_sequences=False)), name="features_embedding_td_2"
        )(features_embedding)

        features_embedding_out = Dropout(config.dropout)(features_embedding_bd)

        # length of sequence not used by the model, but used by the training scorer
        length_input = Input(batch_shape=(None, 1), dtype="int32", name="length_input")

        # combine characters, features and word embeddings
        x = Concatenate()([word_input, chars, features_embedding_out])
        x = Dropout(config.dropout)(x)

        x = Bidirectional(
            LSTM(units=config.num_word_lstm_units, return_sequences=True, recurrent_dropout=config.recurrent_dropout)
        )(x)
        x = Dropout(config.dropout)(x)
        x = Dense(config.num_word_lstm_units, activation="tanh")(x)

        base_model = Model(inputs=[word_input, char_input, features_input, length_input], outputs=[x])
        self.model = CRFModelWrapperDefault(base_model, ntags)
        self.model.build(
            input_shape=[
                (None, None, config.word_embedding_size),
                (None, None, config.max_char_length),
                (None, None, len(config.features_indices)),
                (None, None, 1),
            ]
        )
        self.config = config


class BidLSTM_ChainCRF_FEATURES(BaseModel):
    """
    A Keras implementation of BidLSTM-CRF for sequence labelling using tokens combined with
    additional generic discrete features information and with an alternative CRF layer implementation.
    """

    name = "BidLSTM_ChainCRF_FEATURES"

    def __init__(self, config, ntags=None):
        super().__init__(config, ntags)

        # build input, directly feed with word embedding by the data generator
        word_input = Input(shape=(None, config.word_embedding_size), name="word_input")

        # build character based embedding
        char_input = Input(shape=(None, config.max_char_length), dtype="int32", name="char_input")
        char_embeddings = TimeDistributed(
            Embedding(
                input_dim=config.char_vocab_size,
                output_dim=config.char_embedding_size,
                mask_zero=False,
                # embeddings_initializer=RandomUniform(minval=-0.5, maxval=0.5),
                name="char_embeddings",
            )
        )(char_input)

        chars = TimeDistributed(Bidirectional(LSTM(config.num_char_lstm_units, return_sequences=False)))(
            char_embeddings
        )

        # layout features input and embeddings
        features_input = Input(shape=(None, len(config.features_indices)), dtype="float32", name="features_input")

        # The input dimension is calculated by
        # features_vocabulary_size (default 12) * number_of_features + 1 (the zero is reserved for masking / padding)
        features_embedding = TimeDistributed(
            Embedding(
                input_dim=config.features_vocabulary_size * len(config.features_indices) + 1,
                output_dim=config.features_embedding_size,
                # mask_zero=True,
                trainable=True,
                name="features_embedding",
            ),
            name="features_embedding_td",
        )(features_input)

        features_embedding_bd = TimeDistributed(
            Bidirectional(LSTM(config.features_lstm_units, return_sequences=False)), name="features_embedding_td_2"
        )(features_embedding)

        features_embedding_out = Dropout(config.dropout)(features_embedding_bd)

        # length of sequence not used by the model, but used by the training scorer
        length_input = Input(batch_shape=(None, 1), dtype="int32", name="length_input")

        # combine characters, features and word embeddings
        x = Concatenate()([word_input, chars, features_embedding_out])
        x = Dropout(config.dropout)(x)

        x = Bidirectional(
            LSTM(units=config.num_word_lstm_units, return_sequences=True, recurrent_dropout=config.recurrent_dropout)
        )(x)
        x = Dropout(config.dropout)(x)
        x = Dense(config.num_word_lstm_units, activation="tanh")(x)
        x = Dense(ntags)(x)
        self.crf = ChainCRF()
        pred = self.crf(x)

        self.model = Model(inputs=[word_input, char_input, features_input, length_input], outputs=[pred])
        self.config = config


class BERT(BaseModel):
    """
    A Keras implementation of BERT for sequence labelling with softmax activation final layer.

    For training, the BERT layer will be loaded with weights of existing pre-trained BERT model given by the
    field transformer of the model config (load_pretrained_weights=True).

    For an existing trained model, the BERT layer will be simply initialized (load_pretrained_weights=False),
    without loading pre-trained weights (the weight of the transformer layer will be loaded with the full Keras
    saved model).

    When initializing the model, we can provide a local_path to load locally the transformer config and (if
    necessary) the transformer weights. If local_path=None, these files will be fetched from HuggingFace Hub.
    """

    name = "BERT"

    def __init__(
        self, config, ntags=None, load_pretrained_weights: bool = True, local_path: str = None, preprocessor=None
    ):
        super().__init__(config, ntags, load_pretrained_weights, local_path)

        transformer_layers = self.init_transformer(config, load_pretrained_weights, local_path, preprocessor)

        input_ids_in = Input(shape=(None,), name="input_token", dtype="int32")
        token_type_ids = Input(shape=(None,), name="input_token_type", dtype="int32")
        attention_mask = Input(shape=(None,), name="input_attention_mask", dtype="int32")

        # embedding_layer = transformer_model(input_ids_in, token_type_ids=token_type_ids)[0]
        embedding_layer = transformer_layers(
            input_ids_in, token_type_ids=token_type_ids, attention_mask=attention_mask
        )[0]
        embedding_layer = Dropout(0.1)(embedding_layer)
        label_logits = Dense(ntags, activation="softmax")(embedding_layer)

        self.model = Model(inputs=[input_ids_in, token_type_ids, attention_mask], outputs=[label_logits])
        self.config = config

    def get_generator(self):
        return DataGeneratorTransformers


class BERT_FEATURES(BaseModel):
    """
    A Keras implementation of BERT for sequence labelling combined with additional generic discrete features
    information and with softmax activation final layer.

    For training, the BERT layer will be loaded with weights of existing pre-trained BERT model given by the
    field transformer of the model config (load_pretrained_weights=True).

    For an existing trained model, the BERT layer will be simply initialized (load_pretrained_weights=False),
    without loading pre-trained weights (the weight of the transformer layer will be loaded with the full Keras
    saved model).

    When initializing the model, we can provide a local_path to load locally the transformer config and (if
    necessary) the transformer weights. If local_path=None, these files will be fetched from HuggingFace Hub.
    """

    name = "BERT_FEATURES"

    def __init__(
        self, config, ntags=None, load_pretrained_weights: bool = True, local_path: str = None, preprocessor=None
    ):
        super().__init__(config, ntags, load_pretrained_weights, local_path)

        transformer_layers = self.init_transformer(config, load_pretrained_weights, local_path, preprocessor)

        input_ids_in = Input(shape=(None,), name="input_token", dtype="int32")
        token_type_ids = Input(shape=(None,), name="input_token_type", dtype="int32")
        attention_mask = Input(shape=(None,), name="input_attention_mask", dtype="int32")

        text_embedding_layer = transformer_layers(
            input_ids_in, token_type_ids=token_type_ids, attention_mask=attention_mask
        )[0]
        text_embedding_layer = Dropout(0.1)(text_embedding_layer)

        # layout features input and embeddings
        features_input = Input(shape=(None, len(config.features_indices)), dtype="float32", name="features_input")

        # The input dimension is calculated by
        # features_vocabulary_size (default 12) * number_of_features + 1 (the zero is reserved for masking / padding)
        features_embedding = TimeDistributed(
            Embedding(
                input_dim=config.features_vocabulary_size * len(config.features_indices) + 1,
                output_dim=config.features_embedding_size,
                # mask_zero=True,
                trainable=True,
                name="features_embedding",
            ),
            name="features_embedding_td",
        )(features_input)

        features_embedding_bd = TimeDistributed(
            Bidirectional(LSTM(config.features_lstm_units, return_sequences=False)), name="features_embedding_td_2"
        )(features_embedding)

        features_embedding_out = Dropout(config.dropout)(features_embedding_bd)

        # combine feature and text embeddings
        x = Concatenate()([text_embedding_layer, features_embedding_out])
        x = Dropout(config.dropout)(x)

        x = Bidirectional(
            LSTM(units=config.num_word_lstm_units, return_sequences=True, recurrent_dropout=config.recurrent_dropout)
        )(x)
        x = Dropout(config.dropout)(x)
        label_logits = Dense(ntags, activation="softmax")(x)

        self.model = Model(
            inputs=[input_ids_in, features_input, token_type_ids, attention_mask], outputs=[label_logits]
        )
        self.config = config

    def get_generator(self):
        return DataGeneratorTransformers


class BERT_CRF(BaseModel):
    """
    A Keras implementation of BERT-CRF for sequence labelling. The BERT layer will be loaded with weights
    of existing pre-trained BERT model given by the field transformer in the config.
    """

    name = "BERT_CRF"

    def __init__(
        self,
        config: ModelConfig,
        ntags=None,
        load_pretrained_weights: bool = True,
        local_path: str = None,
        preprocessor=None,
    ):
        super().__init__(config, ntags, load_pretrained_weights, local_path=local_path)

        transformer_layers = self.init_transformer(config, load_pretrained_weights, local_path, preprocessor)

        input_ids_in = Input(shape=(None,), name="input_token", dtype="int32")
        token_type_ids = Input(shape=(None,), name="input_token_type", dtype="int32")
        attention_mask = Input(shape=(None,), name="input_attention_mask", dtype="int32")

        # embedding_layer = transformer_layers(input_ids_in, token_type_ids=token_type_ids)[0]
        embedding_layer = transformer_layers(
            input_ids_in, token_type_ids=token_type_ids, attention_mask=attention_mask
        )[0]
        x = Dropout(0.1)(embedding_layer)

        base_model = Model(inputs=[input_ids_in, token_type_ids, attention_mask], outputs=[x])

        self.model = CRFModelWrapperForBERT(base_model, ntags)
        self.model.build(
            input_shape=[
                (
                    None,
                    None,
                ),
                (
                    None,
                    None,
                ),
                (
                    None,
                    None,
                ),
            ]
        )
        self.config = config

    def get_generator(self):
        return DataGeneratorTransformers


class BERT_ChainCRF(BaseModel):
    """
    A Keras implementation of BERT-CRF for sequence labelling. The BERT layer will be loaded with weights
    of existing pre-trained BERT model given by the field transformer in the config.

    This architecture uses an alternative CRF layer implementation.
    """

    name = "BERT_ChainCRF"

    def __init__(
        self,
        config: ModelConfig,
        ntags=None,
        load_pretrained_weights: bool = True,
        local_path: str = None,
        preprocessor=None,
    ):
        super().__init__(config, ntags, load_pretrained_weights, local_path=local_path)

        transformer_layers = self.init_transformer(config, load_pretrained_weights, local_path, preprocessor)

        input_ids_in = Input(shape=(None,), name="input_token", dtype="int32")
        token_type_ids = Input(shape=(None,), name="input_token_type", dtype="int32")
        attention_mask = Input(shape=(None,), name="input_attention_mask", dtype="int32")

        # embedding_layer = transformer_layers(input_ids_in, token_type_ids=token_type_ids)[0]
        embedding_layer = transformer_layers(
            input_ids_in, token_type_ids=token_type_ids, attention_mask=attention_mask
        )[0]
        embedding_layer = Dropout(0.1)(embedding_layer)
        x = Dense(ntags)(embedding_layer)
        self.crf = ChainCRF()
        pred = self.crf(x)

        self.model = Model(inputs=[input_ids_in, token_type_ids, attention_mask], outputs=[pred])
        self.config = config

    def get_generator(self):
        return DataGeneratorTransformers


class BERT_CRF_FEATURES(BaseModel):
    """
    A Keras implementation of BERT-CRF for sequence labelling using tokens combined with
    additional generic discrete features information. The BERT layer will be loaded with weights
    of existing pre-trained BERT model given by the field transformer in the config.
    """

    name = "BERT_CRF_FEATURES"

    def __init__(self, config, ntags=None, load_pretrained_weights=True, local_path: str = None, preprocessor=None):
        super().__init__(config, ntags, load_pretrained_weights, local_path=local_path)

        transformer_layers = self.init_transformer(config, load_pretrained_weights, local_path, preprocessor)

        input_ids_in = Input(shape=(None,), name="input_token", dtype="int32")
        token_type_ids = Input(shape=(None,), name="input_token_type", dtype="int32")
        attention_mask = Input(shape=(None,), name="input_attention_mask", dtype="int32")

        # text_embedding_layer = transformer_layers(input_ids_in, token_type_ids=token_type_ids)[0]
        text_embedding_layer = transformer_layers(
            input_ids_in, token_type_ids=token_type_ids, attention_mask=attention_mask
        )[0]
        text_embedding_layer = Dropout(0.1)(text_embedding_layer)

        # layout features input and embeddings
        features_input = Input(shape=(None, len(config.features_indices)), dtype="float32", name="features_input")

        # The input dimension is calculated by
        # features_vocabulary_size (default 12) * number_of_features + 1 (the zero is reserved for masking / padding)
        features_embedding = TimeDistributed(
            Embedding(
                input_dim=config.features_vocabulary_size * len(config.features_indices) + 1,
                output_dim=config.features_embedding_size,
                # mask_zero=True,
                trainable=True,
                name="features_embedding",
            ),
            name="features_embedding_td",
        )(features_input)

        features_embedding_bd = TimeDistributed(
            Bidirectional(LSTM(config.features_lstm_units, return_sequences=False)), name="features_embedding_td_2"
        )(features_embedding)

        features_embedding_out = Dropout(config.dropout)(features_embedding_bd)

        # combine feature and text embeddings
        x = Concatenate()([text_embedding_layer, features_embedding_out])
        x = Dropout(config.dropout)(x)

        x = Bidirectional(
            LSTM(units=config.num_word_lstm_units, return_sequences=True, recurrent_dropout=config.recurrent_dropout)
        )(x)
        x = Dropout(config.dropout)(x)
        x = Dense(config.num_word_lstm_units, activation="tanh")(x)

        base_model = Model(inputs=[input_ids_in, features_input, token_type_ids, attention_mask], outputs=[x])

        self.model = CRFModelWrapperForBERT(base_model, ntags)
        self.model.build(
            input_shape=[
                (
                    None,
                    None,
                ),
                (None, None, len(config.features_indices)),
                (
                    None,
                    None,
                ),
                (
                    None,
                    None,
                ),
            ]
        )
        self.config = config

    def get_generator(self):
        return DataGeneratorTransformers


class BERT_ChainCRF_FEATURES(BaseModel):
    """
    A Keras implementation of BERT-CRF for sequence labelling using tokens combined with
    additional generic discrete features information. The BERT layer will be loaded with weights
    of existing pre-trained BERT model given by the field transformer in the config.

    This architecture uses an alternative CRF layer implementation.
    """

    name = "BERT_ChainCRF_FEATURES"

    def __init__(self, config, ntags=None, load_pretrained_weights=True, local_path: str = None, preprocessor=None):
        super().__init__(config, ntags, load_pretrained_weights, local_path=local_path)

        transformer_layers = self.init_transformer(config, load_pretrained_weights, local_path, preprocessor)

        input_ids_in = Input(shape=(None,), name="input_token", dtype="int32")
        token_type_ids = Input(shape=(None,), name="input_token_type", dtype="int32")
        attention_mask = Input(shape=(None,), name="input_attention_mask", dtype="int32")

        # text_embedding_layer = transformer_layers(input_ids_in, token_type_ids=token_type_ids)[0]
        text_embedding_layer = transformer_layers(
            input_ids_in, token_type_ids=token_type_ids, attention_mask=attention_mask
        )[0]
        text_embedding_layer = Dropout(0.1)(text_embedding_layer)

        # layout features input and embeddings
        features_input = Input(shape=(None, len(config.features_indices)), dtype="float32", name="features_input")

        # The input dimension is calculated by
        # features_vocabulary_size (default 12) * number_of_features + 1 (the zero is reserved for masking / padding)
        features_embedding = TimeDistributed(
            Embedding(
                input_dim=config.features_vocabulary_size * len(config.features_indices) + 1,
                output_dim=config.features_embedding_size,
                # mask_zero=True,
                trainable=True,
                name="features_embedding",
            ),
            name="features_embedding_td",
        )(features_input)

        features_embedding_bd = TimeDistributed(
            Bidirectional(LSTM(config.features_lstm_units, return_sequences=False)), name="features_embedding_td_2"
        )(features_embedding)

        features_embedding_out = Dropout(config.dropout)(features_embedding_bd)

        # combine feature and text embeddings
        x = Concatenate()([text_embedding_layer, features_embedding_out])
        x = Dropout(config.dropout)(x)

        x = Bidirectional(
            LSTM(units=config.num_word_lstm_units, return_sequences=True, recurrent_dropout=config.recurrent_dropout)
        )(x)
        x = Dropout(config.dropout)(x)
        x = Dense(config.num_word_lstm_units, activation="tanh")(x)

        x = Dense(ntags)(x)
        self.crf = ChainCRF()
        pred = self.crf(x)

        self.model = Model(inputs=[input_ids_in, features_input, token_type_ids, attention_mask], outputs=[pred])
        self.config = config

    def get_generator(self):
        return DataGeneratorTransformers


class BERT_CRF_CHAR(BaseModel):
    """
    A Keras implementation of BERT-CRF for sequence labelling using tokens combined with
    a character input channel. The BERT layer will be loaded with weights of existing
    pre-trained BERT model given by the field transformer in the config.
    """

    name = "BERT_CRF_CHAR"

    def __init__(self, config, ntags=None, load_pretrained_weights=True, local_path: str = None, preprocessor=None):
        super().__init__(config, ntags, load_pretrained_weights, local_path=local_path)

        transformer_layers = self.init_transformer(config, load_pretrained_weights, local_path, preprocessor)

        input_ids_in = Input(shape=(None,), name="input_token", dtype="int32")
        token_type_ids = Input(shape=(None,), name="input_token_type", dtype="int32")
        attention_mask = Input(shape=(None,), name="input_attention_mask", dtype="int32")

        # text_embedding_layer = transformer_layers(input_ids_in, token_type_ids=token_type_ids)[0]
        text_embedding_layer = transformer_layers(
            input_ids_in, token_type_ids=token_type_ids, attention_mask=attention_mask
        )[0]
        text_embedding_layer = Dropout(0.1)(text_embedding_layer)

        # build character based embedding
        char_input = Input(shape=(None, config.max_char_length), dtype="int32", name="char_input")
        char_embeddings = TimeDistributed(
            Embedding(
                input_dim=config.char_vocab_size,
                output_dim=config.char_embedding_size,
                # mask_zero=True,
                trainable=True,
                # embeddings_initializer=RandomUniform(minval=-0.5, maxval=0.5),
                name="char_embeddings",
            )
        )(char_input)

        chars = TimeDistributed(
            Bidirectional(LSTM(config.num_char_lstm_units, return_sequences=False)), name="chars_rnn"
        )(char_embeddings)

        # combine characters and word embeddings
        x = Concatenate()([text_embedding_layer, chars])
        x = Dropout(config.dropout)(x)

        x = Bidirectional(
            LSTM(units=config.num_word_lstm_units, return_sequences=True, recurrent_dropout=config.recurrent_dropout)
        )(x)
        x = Dropout(config.dropout)(x)
        x = Dense(config.num_word_lstm_units, activation="tanh")(x)

        base_model = Model(inputs=[input_ids_in, char_input, token_type_ids, attention_mask], outputs=[x])
        self.model = CRFModelWrapperForBERT(base_model, ntags)
        self.model.build(
            input_shape=[
                (
                    None,
                    None,
                ),
                (None, None, config.max_char_length),
                (
                    None,
                    None,
                ),
                (
                    None,
                    None,
                ),
            ]
        )
        self.config = config

    def get_generator(self):
        return DataGeneratorTransformers


class BERT_CRF_CHAR_FEATURES(BaseModel):
    """
    A Keras implementation of BERT-CRF for sequence labelling using tokens combined with
    additional generic discrete features information and a character input channel. The
    BERT layer will be loaded with weights of existing pre-trained BERT model given by
    the field transformer in the config.
    """

    name = "BERT_CRF_CHAR_FEATURES"

    def __init__(self, config, ntags=None, load_pretrained_weights=True, local_path: str = None, preprocessor=None):
        super().__init__(config, ntags, load_pretrained_weights, local_path=local_path)

        transformer_layers = self.init_transformer(config, load_pretrained_weights, local_path, preprocessor)

        input_ids_in = Input(shape=(None,), name="input_token", dtype="int32")
        token_type_ids = Input(shape=(None,), name="input_token_type", dtype="int32")
        attention_mask = Input(shape=(None,), name="input_attention_mask", dtype="int32")

        # text_embedding_layer = transformer_layers(input_ids_in, token_type_ids=token_type_ids)[0]
        text_embedding_layer = transformer_layers(
            input_ids_in, token_type_ids=token_type_ids, attention_mask=attention_mask
        )[0]
        text_embedding_layer = Dropout(0.1)(text_embedding_layer)

        # build character based embedding
        char_input = Input(shape=(None, config.max_char_length), dtype="int32", name="char_input")
        char_embeddings = TimeDistributed(
            Embedding(
                input_dim=config.char_vocab_size,
                output_dim=config.char_embedding_size,
                # mask_zero=True,
                trainable=True,
                # embeddings_initializer=RandomUniform(minval=-0.5, maxval=0.5),
                name="char_embeddings",
            )
        )(char_input)

        chars = TimeDistributed(
            Bidirectional(LSTM(config.num_char_lstm_units, return_sequences=False)), name="chars_rnn"
        )(char_embeddings)

        # layout features input and embeddings
        features_input = Input(shape=(None, len(config.features_indices)), dtype="float32", name="features_input")

        # The input dimension is calculated by
        # features_vocabulary_size (default 12) * number_of_features + 1 (the zero is reserved for masking / padding)
        features_embedding = TimeDistributed(
            Embedding(
                input_dim=config.features_vocabulary_size * len(config.features_indices) + 1,
                output_dim=config.features_embedding_size,
                # mask_zero=True,
                trainable=True,
                name="features_embedding",
            ),
            name="features_embedding_td",
        )(features_input)

        features_embedding_bd = TimeDistributed(
            Bidirectional(LSTM(config.features_lstm_units, return_sequences=False)), name="features_embedding_td_2"
        )(features_embedding)

        features_embedding_out = Dropout(config.dropout)(features_embedding_bd)

        # combine feature, characters and word embeddings
        x = Concatenate()([text_embedding_layer, chars, features_embedding_out])
        x = Dropout(config.dropout)(x)

        x = Bidirectional(
            LSTM(units=config.num_word_lstm_units, return_sequences=True, recurrent_dropout=config.recurrent_dropout)
        )(x)
        x = Dropout(config.dropout)(x)
        x = Dense(config.num_word_lstm_units, activation="tanh")(x)

        base_model = Model(
            inputs=[input_ids_in, char_input, features_input, token_type_ids, attention_mask], outputs=[x]
        )
        self.model = CRFModelWrapperForBERT(base_model, ntags)
        self.model.build(
            input_shape=[
                (
                    None,
                    None,
                ),
                (None, None, config.max_char_length),
                (None, None, len(config.features_indices)),
                (
                    None,
                    None,
                ),
                (
                    None,
                    None,
                ),
            ]
        )
        self.config = config

    def get_generator(self):
        return DataGeneratorTransformers
